import json
from urllib import request

workflow_api_path = "./workflow_api.json"


def queue_prompt(prompt):
    p = {"prompt": prompt}
    data = json.dumps(p).encode("utf-8")
    req = request.Request(
        "http://127.0.0.1:9999/prompt", data=data
    )  # comfyui start port
    request.urlopen(req)


with open(workflow_api_path, "r") as fp:
    prompt = json.load(fp)


def generate_texts(min_length=50, max_length=302):
    # 50 world
    base_text = "a female character with long, flowing hair that appears to be made of ethereal, swirling patterns resembling the Northern Lights or Aurora Borealis. The background is dominated by deep blues and purples, creating a mysterious and dramatic atmosphere. The character's face is serene, with pale skin and striking features. She"

    # Additional words pool
    additional_words = [
        "gracefully",
        "beautifully",
        "elegant",
        "radiant",
        "mysteriously",
        "vibrant",
        "softly",
        "gently",
        "luminescent",
        "sparkling",
        "delicately",
        "glowing",
        "brightly",
        "shimmering",
        "enchanting",
        "gloriously",
        "magnificent",
        "majestic",
        "fantastically",
        "dazzlingly",
    ]
    for i in range(min_length, max_length):
        idx = i % len(additional_words)
        base_text = base_text + " " + additional_words[idx]
        yield base_text


generated_texts = list(generate_texts(max_length=101))
generated_texts.reverse()

cout = 0
dimensions = [
    (1024, 1024),
    (1024, 768),
    (1024, 576),
    (1024, 512),
    (512, 1024),
    (768, 512),
    (512, 512),
]

for width, height in dimensions:
    # Set the width and height in the prompt
    prompt["135"]["inputs"]["width"] = width
    prompt["135"]["inputs"]["height"] = height

    # Loop through each generated text and send the prompt to the server
    for text in generated_texts:
        prompt["6"]["inputs"]["text"] = text
        queue_prompt(prompt)
        print(f"{cout=}")
        cout += 1
        break



from functools import singledispatchmethod
from typing import Optional

import torch
from comfy import model_management
from comfy.controlnet import ControlLora, ControlNet
from comfy.model_patcher import ModelPatcher
from comfy.sd import VAE
from onediff.infer_compiler import oneflow_compile
from onediff.infer_compiler.backends.oneflow import (
    OneflowDeployableModule as DeployableModule,
)

from ..booster_interface import BoosterExecutor
from .onediff_controlnet import OneDiffControlLora
from .utils.booster_utils import (
    get_model_type,
    is_fp16_model,
    set_compiled_options,
    set_environment_for_svd_img2vid,
)
from .utils.graph_path import generate_graph_path


class BasicOneFlowBoosterExecutor(BoosterExecutor):
    @singledispatchmethod
    def execute(self, model, ckpt_name=None, **kwargs):
        raise NotImplementedError(f"Cannot execute {type(model)=}")

    @execute.register
    def _(self, model: ModelPatcher, ckpt_name: Optional[str] = None, **kwargs):
        torch_model = model.model.diffusion_model
        if isinstance(torch_model, DeployableModule):
            return model

        set_environment_for_svd_img2vid(model)

        if not is_fp16_model(torch_model):
            print(
                f"Warning: Model {type(torch_model)} is not an FP16 model. Compilation will be skipped!"
            )
            return model

        compiled_model = oneflow_compile(torch_model)

        model.model.diffusion_model = compiled_model

        graph_file = generate_graph_path(f"{type(model).__name__}", model=model.model)
        set_compiled_options(compiled_model, graph_file)

        model.weight_inplace_update = True
        return model

    @execute.register(ControlNet)
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs) -> ControlNet:
        torch_model = model.control_model
        if isinstance(torch_model, DeployableModule):
            return model

        if not is_fp16_model(torch_model):
            type_set = get_model_type(torch_model)
            print(
                f"Warning: Model {type(torch_model)} with parameter types {type_set} is not an FP16 model. Compilation will be skipped!"
            )
            return model

        compiled_model = oneflow_compile(torch_model)
        model.control_model = compiled_model

        graph_file = generate_graph_path(ckpt_name, torch_model)
        set_compiled_options(compiled_model, graph_file)
        return model

    @execute.register(VAE)
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs) -> VAE:
        torch_model = model.first_stage_model
        if isinstance(torch_model, DeployableModule):
            return model

        device = model_management.get_torch_device()
        gpu_name = torch.cuda.get_device_name(device)

        if gpu_name == "NVIDIA A800-SXM4-80GB":
            # TODO Record the problem to issues
            model_type = type(torch_model)
            print(
                f"Warning: Model {model_type} not supported on {gpu_name}. Compilation skipped!"
            )
            return model

        compiled_model = oneflow_compile(torch_model)
        model.first_stage_model = compiled_model

        graph_file = generate_graph_path(ckpt_name, torch_model)
        set_compiled_options(compiled_model, graph_file)
        return model

    @execute.register(ControlLora)
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs):
        def gen_compile_options(model):

            graph_file = generate_graph_path(ckpt_name, model)
            return {
                "graph_file": graph_file,
                "graph_file_device": model_management.get_torch_device(),
            }

        controlnet = OneDiffControlLora.from_controllora(
            model, gen_compile_options=gen_compile_options
        )

        return controlnet



# ComfyUI/custom_nodes/ComfyUI-AnimateDiff-Evolved/animatediff/utils_motion.py
import oneflow as torch
from onediff.infer_compiler.backends.oneflow.transform import register

from ._config import animatediff_of, animatediff_pt

# ComfyUI/custom_nodes/ComfyUI-AnimateDiff-Evolved/animatediff/utils_motion.py
CrossAttentionMM_OF_CLS = animatediff_of.animatediff.utils_motion.CrossAttentionMM
CrossAttentionMM_PT_CLS = animatediff_pt.animatediff.utils_motion.CrossAttentionMM


def exists(val):
    return val is not None


def default(val, d):
    if exists(val):
        return val
    return d


def attention_pytorch(q, k, v, heads, mask=None):
    b, _, dim_head = q.shape
    dim_head //= heads
    # q, k, v = map(
    #     lambda t: t.view(b, -1, heads, dim_head).transpose(1, 2),
    #     (q, k, v),
    # )

    # out = torch.nn.functional.scaled_dot_product_attention(q, k, v, attn_mask=mask, dropout_p=0.0, is_causal=False)
    # out = (
    #     out.transpose(1, 2).reshape(b, -1, heads * dim_head)
    # )
    assert mask is None
    out = torch._C.fused_multi_head_attention_inference_v2(
        query=q,
        query_layout="BM(HK)",
        query_head_size=dim_head,
        key=k,
        key_layout="BM(HK)",
        value=v,
        value_layout="BM(HK)",
        output_layout="BM(HK)",
        causal=False,
    )

    return out


class CrossAttentionMM_OF(CrossAttentionMM_OF_CLS):
    def forward(self, x, context=None, value=None, mask=None, scale_mask=None):
        q = self.to_q(x)
        context = default(context, x)
        k = self.to_k(context)
        if value is not None:
            v = self.to_v(value)
            del value
        else:
            v = self.to_v(context)

        # apply custom scale by multiplying k by scale factor
        if self.scale is not None:
            k *= self.scale

        # apply scale mask, if present
        if scale_mask is not None:
            k *= scale_mask

        # out = optimized_attention_mm(q, k, v, self.heads, mask)
        out = attention_pytorch(q, k, v, self.heads, mask)

        return self.to_out(out)


register(torch2oflow_class_map={CrossAttentionMM_PT_CLS: CrossAttentionMM_OF})



import os
import re
import time

from onediff.infer_compiler import DeployableModule

from .model_patcher import (
    OneFlowDeepCacheSpeedUpModelPatcher,
    OneFlowSpeedUpModelPatcher,
)
from .onediff_load_utils import onediff_load_quant_checkpoint_advanced
from .onediff_quant_utils import (
    quantize_and_save_model,
    replace_module_with_quantizable_module,
)

OUTPUT_FOLDER = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "..", "graphs"
)


def load_graph(deploy_module, graph_filename, device, subfolder):
    graph_filename = os.path.join(OUTPUT_FOLDER, subfolder, graph_filename)
    if isinstance(deploy_module, DeployableModule):
        deploy_module = deploy_module.to(device)
        start_time = time.time()
        if os.path.isfile(graph_filename):
            if deploy_module.get_graph().is_compiled:
                return

            print(f"Loading existed graph from {graph_filename} ...")
            deploy_module.load_graph(graph_filename)
            print(
                f"Graph has been loaded. {time.time() - start_time:.1f} seconds elapsed"
            )
        else:
            print(f"Invalid file to load: {graph_filename}")
    else:
        raise NotImplementedError(f"Unsupported: {type(deploy_module)}")


def save_graph(deploy_module, prefix: str, device: str, subfolder: str):
    str_cls_type = str(type(deploy_module._deployable_module_model._torch_module))
    match = re.search(r"<class '(.+?)'>", str_cls_type)
    module_class_name = match.group(1)

    graph_filename = os.path.join(
        OUTPUT_FOLDER,
        subfolder,
        f"{prefix}-{device}-{module_class_name}.graph",
    )

    if isinstance(deploy_module, DeployableModule):
        deploy_module = deploy_module.to(device)

        start_time = time.time()
        if not os.path.exists(graph_filename):
            if not hasattr(deploy_module.get_graph(), "inputs_original"):
                print(
                    f"Warning: The Graph loaded from the disk doesn't support saving, nor is it necessary to save. Skip saving."
                )
                return
            print(f"Graph saving...")
            deploy_module.save_graph(graph_filename)
            print(
                f"Graph saved to {graph_filename}. {time.time() - start_time:.1f} seconds elapsed"
            )
        else:
            print(
                f"File existed, please change the prefix if you want to save it: {graph_filename}"
            )
    else:
        raise NotImplementedError(f"Unsupported: {type(deploy_module)}")



"""
Torch run example: python examples/text_to_image_deep_cache_sdxl.py --compile 0
Compile to oneflow graph example: python examples/text_to_image_deep_cache_sdxl.py
"""
import argparse
import os

import torch

from onediffx import compile_pipe
from onediffx.deep_cache import StableDiffusionXLPipeline

parser = argparse.ArgumentParser()
parser.add_argument(
    "--base", type=str, default="stabilityai/stable-diffusion-xl-base-1.0"
)
parser.add_argument("--variant", type=str, default="fp16")
parser.add_argument(
    "--prompt",
    type=str,
    default="street style, detailed, raw photo, woman, face, shot on CineStill 800T",
)
parser.add_argument("--height", type=int, default=1024)
parser.add_argument("--width", type=int, default=1024)
parser.add_argument("--n_steps", type=int, default=30)
parser.add_argument("--saved_image", type=str, required=False, default="sdxl-out.png")
parser.add_argument("--warmup", type=int, default=1)
parser.add_argument("--seed", type=int, default=1)
parser.add_argument(
    "--compile",
    type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
    default=True,
)
parser.add_argument(
    "--run_multiple_resolutions",
    type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
    default=False,
)
args = parser.parse_args()

# Normal SDXL pipeline init.
OUTPUT_TYPE = "pil"

# SDXL base: StableDiffusionXLPipeline
base = StableDiffusionXLPipeline.from_pretrained(
    args.base,
    torch_dtype=torch.float16,
    variant=args.variant,
    use_safetensors=True,
)
base.to("cuda")

# Compile unet with oneflow
if args.compile:
    print("Compiling unet with oneflow.")
    base = compile_pipe(base)


# Warmup with run
# Will do compilatioin in the first run
print("Warmup with running graphs...")
torch.manual_seed(args.seed)
image = base(
    prompt=args.prompt,
    height=args.height,
    width=args.width,
    num_inference_steps=args.n_steps,
    output_type=OUTPUT_TYPE,
    cache_interval=3,
    cache_layer_id=0,
    cache_block_id=0,
).images

# Normal SDXL run
print("Normal SDXL run...")
torch.manual_seed(args.seed)
image = base(
    prompt=args.prompt,
    height=args.height,
    width=args.width,
    num_inference_steps=args.n_steps,
    output_type=OUTPUT_TYPE,
    cache_interval=3,
    cache_layer_id=0,
    cache_block_id=0,
).images
image[0].save(f"h{args.height}-w{args.width}-{args.saved_image}")

print("Test run with multiple resolutions...")
if args.run_multiple_resolutions:
    sizes = [960, 720, 896, 768]
    if "CI" in os.environ:
        sizes = [360]
    for h in sizes:
        for w in sizes:
            image = base(
                prompt=args.prompt,
                height=h,
                width=w,
                num_inference_steps=args.n_steps,
                output_type=OUTPUT_TYPE,
                cache_interval=3,
                cache_layer_id=0,
                cache_block_id=0,
            ).images



# The code is referenced from gists https://gist.github.com/w-e-w/0f37c04c18e14e4ee1482df5c4eb9f53 on
# Stable-Diffusion-WebUI API wiki https://github.com/AUTOMATIC1111/stable-diffusion-webui/wiki/API

# To use OneDiff to speed up model inference, you only need to
# add `"script_name": "onediff_diffusion_model"` to the data of the request
# And if you are using OneDiff Enterprise, add another
# `"script_args" : [{"0": True}]` field to enable quantization

import base64
import json
import os
import time
import urllib.request
from datetime import datetime

webui_server_url = "http://127.0.0.1:7860"

out_dir = "api_out"
out_dir_t2i = os.path.join(out_dir, "txt2img")
os.makedirs(out_dir_t2i, exist_ok=True)


def timestamp():
    return datetime.fromtimestamp(time.time()).strftime("%Y%m%d-%H%M%S")


def decode_and_save_base64(base64_str, save_path):
    with open(save_path, "wb") as file:
        file.write(base64.b64decode(base64_str))


def call_api(api_endpoint, **payload):
    data = json.dumps(payload).encode("utf-8")
    request = urllib.request.Request(
        f"{webui_server_url}/{api_endpoint}",
        headers={"Content-Type": "application/json"},
        data=data,
    )
    response = urllib.request.urlopen(request)
    return json.loads(response.read().decode("utf-8"))


def call_txt2img_api(**payload):
    response = call_api("sdapi/v1/txt2img", **payload)
    for index, image in enumerate(response.get("images")):
        save_path = os.path.join(out_dir_t2i, f"txt2img-{timestamp()}-{index}.png")
        decode_and_save_base64(image, save_path)


if __name__ == "__main__":
    payload = {
        "prompt": "masterpiece, (best quality:1.1), 1girl",
        # "prompt": "masterpiece, (best quality:1.1), 1girl <lora:lora_model:1>",  # extra networks also in prompts
        "negative_prompt": "",
        "seed": 1,
        "steps": 20,
        "width": 1024,
        "height": 1024,
        "cfg_scale": 7,
        "n_iter": 1,
        "batch_size": 1,
        # Enable OneDiff speed up
        "script_name": "onediff_diffusion_model",
        # If you are using OneDiff Enterprise, add the field below to enable quant feature
        # "script_args" : [{"0": True}],
    }
    call_txt2img_api(**payload)



# The code is referenced from gists https://gist.github.com/w-e-w/0f37c04c18e14e4ee1482df5c4eb9f53 on
# Stable-Diffusion-WebUI API wiki https://github.com/AUTOMATIC1111/stable-diffusion-webui/wiki/API

# To use OneDiff to speed up model inference, you only need to
# add `"script_name": "onediff_diffusion_model"` to the data of the request
# And if you are using OneDiff Enterprise, add another
# `"script_args" : [{"0": True}]` field to enable quantization

import base64
import json
import os
import time
import urllib.request
from datetime import datetime
from pathlib import Path

webui_server_url = "http://127.0.0.1:7860"

out_dir = "api_out"
out_dir_i2i = os.path.join(out_dir, "img2img")
os.makedirs(out_dir_i2i, exist_ok=True)


def timestamp():
    return datetime.fromtimestamp(time.time()).strftime("%Y%m%d-%H%M%S")


def encode_file_to_base64(path):
    with open(path, "rb") as file:
        return base64.b64encode(file.read()).decode("utf-8")


def decode_and_save_base64(base64_str, save_path):
    with open(save_path, "wb") as file:
        file.write(base64.b64decode(base64_str))


def call_api(api_endpoint, **payload):
    data = json.dumps(payload).encode("utf-8")
    request = urllib.request.Request(
        f"{webui_server_url}/{api_endpoint}",
        headers={"Content-Type": "application/json"},
        data=data,
    )
    response = urllib.request.urlopen(request)
    return json.loads(response.read().decode("utf-8"))


def call_img2img_api(**payload):
    response = call_api("sdapi/v1/img2img", **payload)
    for index, image in enumerate(response.get("images")):
        save_path = os.path.join(out_dir_i2i, f"img2img-{timestamp()}-{index}.png")
        decode_and_save_base64(image, save_path)


if __name__ == "__main__":
    img_path = str(Path(__file__).parent / "cat.png")
    init_images = [
        encode_file_to_base64(img_path),
        # encode_file_to_base64(r"B:\path\to\img_2.png"),
        # "https://image.can/also/be/a/http/url.png",
    ]

    # batch_size = 2
    batch_size = 1
    payload = {
        "prompt": "1girl, blue hair",
        "seed": 1,
        "steps": 20,
        "width": 1024,
        "height": 1024,
        "denoising_strength": 0.5,
        "n_iter": 1,
        "init_images": init_images,
        "batch_size": batch_size if len(init_images) == 1 else len(init_images),
        # Enable OneDiff speed up
        "script_name": "onediff_diffusion_model",
        # If you are using OneDiff Enterprise, add the field below to enable quant feature
        # "script_args" : [{"0": True}],
        # "mask": encode_file_to_base64(r"B:\path\to\mask.png")
    }
    # if len(init_images) > 1 then batch_size should be == len(init_images)
    # else if len(init_images) == 1 then batch_size can be any value int >= 1
    call_img2img_api(**payload)

    # there exist a useful extension that allows converting of webui calls to api payload
    # particularly useful when you wish setup arguments of extensions and scripts
    # https://github.com/huchenlei/sd-webui-api-payload-display



"""hijack ComfyUI/comfy/utils.py"""
import torch
from comfy.utils import copy_to_param
from onediff.infer_compiler.backends.oneflow.param_utils import (
    update_graph_related_tensor,
)

from ..sd_hijack_utils import Hijacker


@torch.no_grad()
def copy_to_param_of(org_fn, obj, attr, value):
    # inplace update tensor instead of replacing it
    attrs = attr.split(".")
    for name in attrs[:-1]:
        obj = getattr(obj, name)

    prev = getattr(obj, attrs[-1])

    if prev.data.dtype == torch.int8 and prev.data.dtype != value.dtype:
        return

    prev.data.copy_(value)

    if isinstance(obj, torch.nn.Conv2d):
        update_graph_related_tensor(obj)


def cond_func(orig_func, *args, **kwargs):
    return True


comfy_utils_hijack = Hijacker()

comfy_utils_hijack.register(
    orig_func=copy_to_param, sub_func=copy_to_param_of, cond_func=cond_func
)



import argparse
import json
import time

import nexfort

import torch
from diffusers import FluxPipeline


def parse_args():
    parser = argparse.ArgumentParser(
        description="Use nexfort to accelerate image generation with Flux."
    )
    parser.add_argument(
        "--model",
        type=str,
        default="black-forest-labs/FLUX.1-dev",
        help="Model path or identifier.",
    )
    parser.add_argument(
        "--speedup-t5",
        action="store_true",
        help="Enable optimize t5.",
    )
    parser.add_argument(
        "--quantize",
        action="store_true",
        help="Enable fp8 quantization.",
    )
    parser.add_argument(
        "--transform",
        action="store_true",
        help="Enable speedup with nexfort.",
    )
    parser.add_argument(
        "--prompt",
        type=str,
        default="evening sunset scenery blue sky nature, glass bottle with a galaxy in it",
        help="Prompt for the image generation.",
    )
    parser.add_argument(
        "--height", type=int, default=1024, help="Height of the generated image."
    )
    parser.add_argument(
        "--width", type=int, default=1024, help="Width of the generated image."
    )
    parser.add_argument(
        "--guidance_scale",
        type=float,
        default=0.0,
        help="The scale factor for the guidance.",
    )
    parser.add_argument(
        "--max_sequence_length",
        type=int,
        default=256,
        help="Maximum sequence length to use with the `prompt`.",
    )
    parser.add_argument(
        "--num-inference-steps", type=int, default=20, help="Number of inference steps."
    )
    parser.add_argument(
        "--saved-image",
        type=str,
        default="./flux.png",
        help="Path to save the generated image.",
    )
    parser.add_argument(
        "--seed", type=int, default=1, help="Seed for random number generation."
    )
    parser.add_argument(
        "--run_multiple_resolutions",
        action="store_true",
    )
    parser.add_argument(
        "--run_multiple_prompts",
        action="store_true",
    )
    return parser.parse_args()


args = parse_args()

device = torch.device("cuda")


def generate_texts(min_length=50, max_length=302):
    base_text = "a female character with long, flowing hair that appears to be made of ethereal, swirling patterns resembling the Northern Lights or Aurora Borealis. The background is dominated by deep blues and purples, creating a mysterious and dramatic atmosphere. The character's face is serene, with pale skin and striking features. She"

    additional_words = [
        "gracefully",
        "beautifully",
        "elegant",
        "radiant",
        "mysteriously",
        "vibrant",
        "softly",
        "gently",
        "luminescent",
        "sparkling",
        "delicately",
        "glowing",
        "brightly",
        "shimmering",
        "enchanting",
        "gloriously",
        "magnificent",
        "majestic",
        "fantastically",
        "dazzlingly",
    ]

    for i in range(min_length, max_length):
        idx = i % len(additional_words)
        base_text += " " + additional_words[idx]
        yield base_text


class FluxGenerator:
    def __init__(
        self,
        model,
        enable_quantize=False,
        enable_fast_transformer=False,
        enable_speedup_t5=False,
    ):
        self.pipe = FluxPipeline.from_pretrained(
            model,
            torch_dtype=torch.bfloat16,
        )

        # Put the quantize process after `self.pipe.to(device)` if you have more than 32GB ram.
        if enable_quantize:
            print("quant...")
            from nexfort.quantization import quantize

            self.pipe.transformer = quantize(
                self.pipe.transformer, quant_type="fp8_e4m3_e4m3_dynamic_per_tensor"
            )
            if enable_speedup_t5:
                self.pipe.text_encoder_2 = quantize(
                    self.pipe.text_encoder_2,
                    quant_type="fp8_e4m3_e4m3_dynamic_per_tensor",
                )

        self.pipe.to(device)

        if enable_fast_transformer:
            print("compile...")
            from nexfort.compilers import transform

            self.pipe.transformer = transform(self.pipe.transformer)
            if enable_speedup_t5:
                self.pipe.text_encoder_2 = transform(self.pipe.text_encoder_2)

    def warmup(self, gen_args, warmup_iterations=1):
        warmup_args = gen_args.copy()

        # warmup_args["generator"] = torch.Generator(device=device).manual_seed(0)
        torch.manual_seed(args.seed)

        print("Starting warmup...")
        start_time = time.time()
        for _ in range(warmup_iterations):
            self.pipe(**warmup_args)
        end_time = time.time()
        print("Warmup complete.")
        print(f"Warmup time: {end_time - start_time:.2f} seconds")

    def generate(self, gen_args):
        # gen_args["generator"] = torch.Generator(device=device).manual_seed(args.seed)
        torch.manual_seed(args.seed)

        # Run the model
        start_time = time.time()
        image = self.pipe(**gen_args).images[0]
        end_time = time.time()

        image.save(args.saved_image)

        return image, end_time - start_time


def main():
    flux = FluxGenerator(args.model, args.quantize, args.transform, args.speedup_t5)

    if args.run_multiple_prompts:
        dynamic_prompts = generate_texts(max_length=101)
        prompt_list = list(dynamic_prompts)
    else:
        prompt_list = [args.prompt]

    gen_args = {
        "prompt": args.prompt,
        "num_inference_steps": args.num_inference_steps,
        "height": args.height,
        "width": args.width,
        "guidance_scale": args.guidance_scale,
        "max_sequence_length": args.max_sequence_length,
    }

    flux.warmup(gen_args)

    for prompt in prompt_list:
        gen_args["prompt"] = prompt
        print(f"Processing prompt of length {len(prompt)} characters.")
        image, inference_time = flux.generate(gen_args)
        print(
            f"Generated image saved to {args.saved_image} in {inference_time:.2f} seconds."
        )
        cuda_mem_after_used = torch.cuda.max_memory_allocated() / (1024**3)
        print(f"Max used CUDA memory : {cuda_mem_after_used:.3f} GiB")

    if args.run_multiple_resolutions:
        gen_args["prompt"] = args.prompt
        print("Test run with multiple resolutions...")
        sizes = [1536, 1024, 768, 720, 576, 512, 256]
        for h in sizes:
            for w in sizes:
                gen_args["height"] = h
                gen_args["width"] = w
                print(f"Running at resolution: {h}x{w}")
                start_time = time.time()
                flux.generate(gen_args)
                end_time = time.time()
                print(f"Inference time: {end_time - start_time:.2f} seconds")


if __name__ == "__main__":
    main()



from onediff.infer_compiler.backends.oneflow import online_quantization_utils
from register_comfy.CrossAttentionPatch import is_crossAttention_patch

from .patch_executor import PatchExecutorBase


def convert_to_nested(original_dict):
    new_dict = {}

    for key, value in original_dict.items():
        if isinstance(value, dict):
            new_value = convert_to_nested(value)
        elif is_crossAttention_patch(value):
            new_value = value.get_bind_model()
        else:
            new_value = value

        # Update the new dictionary with the processed value
        new_dict[key] = new_value
    return new_dict


class QuantizedInputPatch(PatchExecutorBase):
    def __init__(self):
        self.is_use_patch = False

    def set_patch(self):
        if self.check_patch():
            return

        def new_patch_input_adapter(in_args, in_kwargs):
            return in_args, convert_to_nested(in_kwargs)

        self.is_use_patch = True
        online_quantization_utils.patch_input_adapter = new_patch_input_adapter

    def get_patch(self, module):
        pass

    def check_patch(self):
        return self.is_use_patch



# /ComfyUI/custom_nodes/ComfyUI-AnimateDiff-Evolved/animatediff/sampling.py
import oneflow as flow  # usort: skip
from einops import rearrange
from onediff.infer_compiler import DeployableModule
from onediff.infer_compiler.backends.oneflow.transform import register
from oneflow.nn.functional import group_norm

from ._config import animatediff_hijacker, animatediff_of, animatediff_pt, comfy_of

FunctionInjectionHolder = animatediff_pt.animatediff.sampling.FunctionInjectionHolder


def cast_bias_weight(s, input):
    bias = None
    # non_blocking = comfy.model_management.device_supports_non_blocking(input.device)
    non_blocking = False
    if s.bias is not None:
        bias = s.bias.to(
            device=input.device, dtype=input.dtype, non_blocking=non_blocking
        )
    weight = s.weight.to(
        device=input.device, dtype=input.dtype, non_blocking=non_blocking
    )
    return weight, bias


def groupnorm_mm_factory(params, manual_cast=False):
    def groupnorm_mm_forward(self, input):
        # axes_factor normalizes batch based on total conds and unconds passed in batch;
        # the conds and unconds per batch can change based on VRAM optimizations that may kick in
        if not params.is_using_sliding_context():
            batched_conds = input.size(0) // params.full_length
        else:
            batched_conds = input.size(0) // params.context_options.context_length

        # input = rearrange(input, "(b f) c h w -> b c f h w", b=batched_conds)
        input = input.unflatten(0, (batched_conds, -1)).permute(0, 2, 1, 3, 4)

        if manual_cast:
            # weight, bias = comfy_of.ops.cast_bias_weight(self, input)
            weight, bias = cast_bias_weight(self, input)
        else:
            weight, bias = self.weight, self.bias
        input = group_norm(input, self.num_groups, weight, bias, self.eps)
        # input = rearrange(input, "b c f h w -> (b f) c h w", b=batched_conds)
        input = input.permute(0, 2, 1, 3, 4).flatten(0, 1)
        return input

    return groupnorm_mm_forward


# ComfyUI/custom_nodes/ComfyUI-AnimateDiff-Evolved/animatediff/motion_module_ad.py
AnimateDiffVersion = animatediff_pt.animatediff.motion_module_ad.AnimateDiffVersion
AnimateDiffFormat = animatediff_pt.animatediff.motion_module_ad.AnimateDiffFormat
# ComfyUI/custom_nodes/ComfyUI-AnimateDiff-Evolved/animatediff/utils_model.py
ModelTypeSD = animatediff_pt.animatediff.utils_model.ModelTypeSD


class Handles:
    def __init__(self):
        self.handles = []

    def add(self, obj, key, value):
        org_attr = getattr(obj, key, None)
        setattr(obj, key, value)
        self.handles.append(lambda: setattr(obj, key, org_attr))

    def restore(self):
        for handle in self.handles:
            handle()
        self.handles = []


handles = Handles()


def inject_functions(orig_func, self, model, params):

    ret = orig_func(self, model, params)

    if model.motion_models is not None:
        # only apply groupnorm hack if not [v3 or ([not Hotshot] and SD1.5 and v2 and apply_v2_properly)]
        info = model.motion_models[0].model.mm_info
        if not (
            info.mm_version == AnimateDiffVersion.V3
            or (
                info.mm_format not in [AnimateDiffFormat.HOTSHOTXL]
                and info.sd_type == ModelTypeSD.SD1_5
                and info.mm_version == AnimateDiffVersion.V2
                and params.apply_v2_properly
            )
        ):

            handles.add(flow.nn.GroupNorm, "forward", groupnorm_mm_factory(params))
            # comfy_of.ops.manual_cast.GroupNorm.forward_comfy_cast_weights = groupnorm_mm_factory(params, manual_cast=True)
            handles.add(
                comfy_of.ops.manual_cast.GroupNorm,
                "forward_comfy_cast_weights",
                groupnorm_mm_factory(params, manual_cast=True),
            )

        del info
    return ret


def restore_functions(orig_func, *args, **kwargs):
    ret = orig_func(*args, **kwargs)
    handles.restore()
    return ret


def cond_func(orig_func, self, model, *args, **kwargs):
    diff_model = model.model.diffusion_model
    if isinstance(diff_model, DeployableModule):
        return True
    else:
        return False


animatediff_hijacker.register(
    FunctionInjectionHolder.inject_functions,
    inject_functions,
    cond_func,
)

animatediff_hijacker.register(
    FunctionInjectionHolder.restore_functions,
    restore_functions,
    cond_func,
)



from enum import Enum


class OneDiffBackend(Enum):
    ONEFLOW = "oneflow"
    NEXFORT = "nexfort"

    def __str__(self):
        return self.value

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}: {self.value}>"



import torch
from diffusers.pipelines.wuerstchen.modeling_wuerstchen_common import (
    AttnBlock,
    TimestepBlock,
    WuerstchenLayerNorm,
)
from diffusers.pipelines.wuerstchen.modeling_wuerstchen_diffnext import ResBlockStageB


def patch_prior_fp16_overflow(prior, num_overflow_up_blocks=1):
    for i in range(len(prior.up_blocks) - num_overflow_up_blocks, len(prior.up_blocks)):
        prior.up_blocks[i].to(torch.bfloat16)
        prior.up_upscalers[i].to(torch.bfloat16)
        prior.up_repeat_mappers[i].to(torch.bfloat16)

    def _up_decode(self, level_outputs, r_embed, clip):
        x = level_outputs[0]
        block_group = zip(self.up_blocks, self.up_upscalers, self.up_repeat_mappers)
        for i, (up_block, upscaler, repmap) in enumerate(block_group):
            if i == len(self.up_blocks) - num_overflow_up_blocks:
                x = x.to(torch.bfloat16)
                r_embed = r_embed.to(torch.bfloat16)
                clip = clip.to(torch.bfloat16)
            for j in range(len(repmap) + 1):
                for k, block in enumerate(up_block):
                    block_class = block.__class__
                    if isinstance(block, ResBlockStageB):
                        skip = level_outputs[i] if k == 0 and i > 0 else None
                        if skip is not None and (
                            x.size(-1) != skip.size(-1) or x.size(-2) != skip.size(-2)
                        ):
                            x = torch.nn.functional.interpolate(
                                x.float(),
                                skip.shape[-2:],
                                mode="bilinear",
                                align_corners=True,
                            )
                        if (
                            skip is not None
                            and i >= len(level_outputs) - num_overflow_up_blocks
                        ):
                            skip = skip.to(torch.bfloat16)
                        x = block(x, skip)
                    elif isinstance(block, AttnBlock):
                        x = block(x, clip)
                    elif isinstance(block, TimestepBlock):
                        x = block(x, r_embed)
                    else:
                        x = block(x)
                if j < len(repmap):
                    x = repmap[j](x)
            x = upscaler(x)
        return x

    prior._up_decode = _up_decode.__get__(prior)

    prior.clf.to(torch.bfloat16)
    prior_clf_forward = prior.clf.forward

    def new_prior_clf_forward(x):
        return prior_clf_forward(x).to(torch.float16)

    prior.clf.forward = new_prior_clf_forward

    return prior


original_pixel_shuffle = torch.nn.functional.pixel_shuffle


def pixel_shuffle(input, upscale_factor):
    # https://blog.csdn.net/ONE_SIX_MIX/article/details/103757856
    # Thanks ChatGPT: https://chat.openai.com/share/1ba80104-406f-4bb8-8292-105f69e0452e

    assert input.dim() >= 3, "Input tensor must have at least 3 dimensions"

    # Separate batch_dims and original C, H, W dimensions
    *batch_dims, channels, height, width = input.shape

    assert (
        channels % (upscale_factor**2) == 0
    ), "Number of channels must be divisible by the square of the upscale factor"

    # Calculate new channels after applying upscale_factor
    new_channels = channels // (upscale_factor**2)

    # Reshape input to (*batch_dims, new_channels, upscale_factor, upscale_factor, height, width)
    reshaped = input.reshape(
        *batch_dims, new_channels, upscale_factor, upscale_factor, height, width
    )

    # Adjust permute to handle dynamic batch dimensions
    permute_dims = list(range(len(batch_dims))) + [
        len(batch_dims),
        len(batch_dims) + 3,
        len(batch_dims) + 1,
        len(batch_dims) + 4,
        len(batch_dims) + 2,
    ]
    permuted = reshaped.permute(*permute_dims)

    # Final reshape to get to the target shape
    output = permuted.reshape(
        *batch_dims, new_channels, height * upscale_factor, width * upscale_factor
    )

    return output


def pixel_unshuffle(input, downscale_factor):
    # Thanks ChatGPT: https://chat.openai.com/share/1ba80104-406f-4bb8-8292-105f69e0452e

    assert input.dim() >= 3, "Input tensor must have at least 3 dimensions"

    # Separate batch_dims and original C, H, W dimensions
    *batch_dims, channels, height, width = input.shape

    # Ensure H and W are divisible by downscale_factor
    assert (
        height % downscale_factor == 0 and width % downscale_factor == 0
    ), "Height and Width must be divisible by the downscale factor"

    # Reshape
    reshaped = input.reshape(
        *batch_dims,
        channels,
        height // downscale_factor,
        downscale_factor,
        width // downscale_factor,
        downscale_factor,
    )

    # Adjust permutation indices for tensors with dimensions > 4
    permute_dims = list(range(len(batch_dims))) + [
        len(batch_dims),
        len(batch_dims) + 2,
        len(batch_dims) + 4,
        len(batch_dims) + 1,
        len(batch_dims) + 3,
    ]
    permuted = reshaped.permute(*permute_dims)

    # Final reshape
    output = permuted.reshape(
        *batch_dims,
        channels * downscale_factor**2,
        height // downscale_factor,
        width // downscale_factor,
    )

    return output


# pixel_shuffle() and pixel_unshuffle() call Tensor.sizes() which is not supported by dynamo.
def patch_torch_compile():
    torch.nn.functional.pixel_shuffle = pixel_shuffle
    torch.nn.functional.pixel_unshuffle = pixel_unshuffle



from ..modules.torch_compile.booster_basic import (  # type: ignore
    TorchCompileBoosterExecutor,
)


class OneDiffTorchCompileBooster:
    @classmethod
    def INPUT_TYPES(s):
        # [docs]https://pytorch.org/docs/stable/_modules/torch.html#compile
        return {
            "required": {
                "fullgraph": ([False, True],),
                "dynamic": ([None, True, False],),
                "backend": (["inductor"],),
                "mode": (
                    [
                        "default",
                        "reduce-overhead",
                        "max-autotune",
                        "max-autotune-no-cudagraphs",
                    ],
                ),
                "disable": ([False, True],),
                "docs_link": (
                    "STRING",
                    {
                        "multiline": True,  # True if you want the field to look like the one on the ClipTextEncode node
                        "default": "[docs]: https://pytorch.org/docs/stable/_modules/torch.html#compile",
                    },
                ),
            }
        }

    CATEGORY = "OneDiff/Booster"
    RETURN_TYPES = ("TorchCompileBooster",)
    FUNCTION = "apply"

    def apply(
        self,
        fullgraph=False,
        dynamic=None,
        backend="inductor",
        mode="default",
        options=None,
        disable=False,
        docs_link=None,
    ):
        return (
            TorchCompileBoosterExecutor(
                fullgraph=fullgraph,
                dynamic=dynamic,
                backend=backend,
                mode=mode,
                options=options,
                disable=disable,
            ),
        )


NODE_CLASS_MAPPINGS = {
    "OneDiffTorchCompileBooster": OneDiffTorchCompileBooster,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "OneDiffTorchCompileBooster": "Torch Compile Booster - OneDiff"
}



import random
from pathlib import Path
from typing import Dict, List, Tuple

import numpy as np
import pytest
import safetensors.torch

import torch
from diffusers import DiffusionPipeline
from onediff.infer_compiler import oneflow_compile

from onediffx.lora import (
    delete_adapters,
    get_active_adapters,
    load_and_fuse_lora,
    load_lora_and_optionally_fuse,
    set_and_fuse_adapters,
    unfuse_lora,
)
from PIL import Image
from skimage.metrics import structural_similarity

HEIGHT = 1024
WIDTH = 1024
NUM_STEPS = 30
LORA_SCALE = 0.5
LATENTS = torch.randn(
    1,
    4,
    128,
    128,
    generator=torch.cuda.manual_seed(0),
    dtype=torch.float16,
    device="cuda",
)

image_file_prefix = "/share_nfs/onediff_ci/diffusers/images/1.0"
if not Path(image_file_prefix).exists():
    Path(image_file_prefix).mkdir(parents=True)


@pytest.fixture
def prepare_loras() -> Dict[str, Dict[str, torch.Tensor]]:
    loras = [
        "/share_nfs/onediff_ci/diffusers/loras/SDXL-Emoji-Lora-r4.safetensors",
        "/share_nfs/onediff_ci/diffusers/loras/sdxl_metal_lora.safetensors",
        "/share_nfs/onediff_ci/diffusers/loras/simple_drawing_xl_b1-000012.safetensors",
        "/share_nfs/onediff_ci/diffusers/loras/texta.safetensors",
        "/share_nfs/onediff_ci/diffusers/loras/watercolor_v1_sdxl_lora.safetensors",
    ]
    loras = {Path(x).stem: safetensors.torch.load_file(x) for x in loras}
    return loras


@pytest.fixture
def get_loras(prepare_loras) -> Dict[str, Dict[str, torch.Tensor]]:
    def _get_loras():
        return {name: lora_dict.copy() for name, lora_dict in prepare_loras.items()}

    return _get_loras


@pytest.fixture
def get_multi_loras(prepare_loras) -> Dict[str, Dict[str, torch.Tensor]]:
    def _get_multi_loras():
        multi_lora = {}
        current_name = []
        current_lora = []
        for name, lora_dict in prepare_loras.items():
            current_name.append(name)
            current_lora.append(lora_dict)
            multi_lora[tuple(current_name)] = current_lora
        return multi_lora

    return _get_multi_loras


def get_pipe(name: str = "sdxl"):
    id_dict = {
        "sd1.5": "/share_nfs/hf_models/stable-diffusion-v1-5",
        "sdxl": "/share_nfs/hf_models/stable-diffusion-xl-base-1.0",
        "sd2.1": "/share_nfs/hf_models/stable-diffusion-2-1",
    }
    MODEL_ID = id_dict[name]
    pipeline = DiffusionPipeline.from_pretrained(
        MODEL_ID, variant="fp16", torch_dtype=torch.float16
    ).to("cuda")
    return pipeline


@pytest.fixture(scope="session")
def pipe():
    pipeline = get_pipe()
    return pipeline


def generate_image(pipe):
    image = pipe(
        "masterpiece, best quality, mountain",
        height=HEIGHT,
        width=WIDTH,
        num_inference_steps=NUM_STEPS,
        generator=torch.manual_seed(0),
        latents=LATENTS.clone(),
    ).images[0]
    return image


def prepare_target_images(pipe, loras):
    target_images_list = [
        f"{image_file_prefix}/test_sdxl_lora_{name}_{HEIGHT}_{WIDTH}.png"
        for name in loras
    ]
    if all(Path(x).exists() for x in target_images_list):
        return

    print("Didn't find target images, try to generate...")
    for name, lora in loras.items():
        pipe.load_lora_weights(lora.copy())
        pipe.fuse_lora()
        image = generate_image(pipe)
        pipe.unfuse_lora()
        pipe.unload_lora_weights()
        image.save(f"{image_file_prefix}/test_sdxl_lora_{name}_{HEIGHT}_{WIDTH}.png")
    torch.cuda.empty_cache()


def prepare_target_images_multi_lora(pipe, loras, multi_loras):
    target_images_list = [
        f"{image_file_prefix}/test_sdxl_multi_lora_{'_'.join(names)}_{HEIGHT}_{WIDTH}.png"
        for names in multi_loras
    ]
    if all(Path(x).exists() for x in target_images_list):
        return

    for name, lora in loras.items():
        print(f"loading name: {name}")
        pipe.load_lora_weights(lora.copy(), adapter_name=name)

    print("Didn't find target images, try to generate...")
    for names, loras in multi_loras.items():
        pipe.set_adapters(
            names,
            [
                LORA_SCALE,
            ]
            * len(names),
        )
        image = generate_image(pipe)
        image_name = f"{image_file_prefix}/test_sdxl_multi_lora_{'_'.join(names)}_{HEIGHT}_{WIDTH}.png"
        image.save(image_name)
    pipe.unload_lora_weights()
    torch.cuda.empty_cache()


def preload_multi_loras(pipe, loras):
    for name, lora in loras.items():
        load_lora_and_optionally_fuse(
            pipe,
            lora.copy(),
            adapter_name=name,
            fuse=False,
        )
        unfuse_lora(pipe)


@pytest.mark.skip(reason="onediffx.lora is deprecated")
def test_lora_loading(pipe, get_loras):
    pipe.unet = oneflow_compile(pipe.unet)
    pipe(
        "a cat",
        height=HEIGHT,
        width=WIDTH,
        num_inference_steps=NUM_STEPS,
    ).images[0]
    loras = get_loras()
    prepare_target_images(pipe, loras)

    for name, lora in loras.items():
        load_and_fuse_lora(pipe, lora.copy())
        images_fusion = generate_image(pipe)
        target_image = np.array(
            Image.open(
                f"{image_file_prefix}/test_sdxl_lora_{name}_{HEIGHT}_{WIDTH}.png"
            )
        )
        curr_image = np.array(images_fusion)
        ssim = structural_similarity(
            curr_image, target_image, channel_axis=-1, data_range=255
        )
        unfuse_lora(pipe)
        images_fusion.save(f"./test_sdxl_lora_{name}_{HEIGHT}_{WIDTH}.png")

        print(f"lora {name} ssim {ssim}")
        assert ssim > 0.92, f"LoRA {name} ssim too low"


@pytest.mark.skip(reason="onediffx.lora is deprecated")
def test_multi_lora_loading(pipe, get_multi_loras, get_loras):
    pipe.unet = oneflow_compile(pipe.unet)
    multi_loras = get_multi_loras()
    loras = get_loras()
    prepare_target_images_multi_lora(pipe, loras, multi_loras)
    preload_multi_loras(pipe, loras)

    for names, loras in multi_loras.items():
        set_and_fuse_adapters(
            pipe,
            names,
            [
                LORA_SCALE,
            ]
            * len(names),
        )

        images_fusion = generate_image(pipe)
        image_name = "_".join(names)
        target_image = np.array(
            Image.open(
                f"{image_file_prefix}/test_sdxl_multi_lora_{image_name}_{HEIGHT}_{WIDTH}.png"
            )
        )
        images_fusion.save(f"./test_sdxl_multi_lora_{image_name}_{HEIGHT}_{WIDTH}.png")
        curr_image = np.array(images_fusion)
        ssim = structural_similarity(
            curr_image, target_image, channel_axis=-1, data_range=255
        )
        print(f"lora {names} ssim {ssim}")
        assert ssim > 0.92, f"LoRA {names} ssim too low"


@pytest.mark.skip(reason="onediffx.lora is deprecated")
def test_get_active_adapters(pipe, get_multi_loras, get_loras):
    multi_loras = get_multi_loras()
    preload_multi_loras(pipe, get_loras())
    for names, _ in multi_loras.items():
        set_and_fuse_adapters(pipe, names)
        active_adapters = get_active_adapters(pipe)
        print(f"current adapters: {active_adapters}, target adapters: {names}")
        assert set(active_adapters) == set(names)


@pytest.mark.skip(reason="onediffx.lora is deprecated")
def test_delete_adapters(pipe, get_multi_loras, get_loras):
    multi_loras = get_multi_loras()
    for names, _ in multi_loras.items():
        preload_multi_loras(pipe, get_loras())
        names_to_delete = random.sample(names, k=random.randint(0, len(names)))
        set_and_fuse_adapters(pipe, names)
        delete_adapters(pipe, names_to_delete)
        active_adapters = get_active_adapters(pipe)
        print(
            f"current adapters: {active_adapters}, target adapters: {list(set(names) - set(names_to_delete))}"
        )
        assert set(active_adapters) == set(names) - set(names_to_delete)


@pytest.mark.skip(reason="onediffx.lora is deprecated")
def test_lora_numerical_stability():
    original_pipe = get_pipe("sd1.5")
    pipe = get_pipe("sd1.5")
    loras = [
        "/share_nfs/onediff_ci/diffusers/loras/SD15-IllusionDiffusionPattern-LoRA.safetensors",
        "/share_nfs/onediff_ci/diffusers/loras/SD15-Megaphone-LoRA.safetensors",
    ]
    loras = {Path(x).stem: safetensors.torch.load_file(x) for x in loras}
    param_name = "down_blocks.0.attentions.0.transformer_blocks.0.attn1.to_q.weight"

    for _ in range(1000):
        for name, lora in loras.items():
            load_lora_and_optionally_fuse(
                pipe, lora.copy(), adapter_name=name, fuse=False
            )
        set_and_fuse_adapters(
            pipe, adapter_names=list(loras.keys()), adapter_weights=[0.2, 0.2]
        )
        for name in loras:
            delete_adapters(pipe, name, safe_delete=True)

    assert torch.allclose(
        original_pipe.unet.get_parameter(param_name),
        pipe.unet.get_parameter(param_name),
        rtol=0,
        atol=1e-3,
    )
    print(
        f"numerical stability: max diff is {(original_pipe.unet.get_parameter(param_name) - pipe.unet.get_parameter(param_name)).abs().max().item()}"
    )



# The code is mainly copied from https://github.com/huggingface/diffusers/blob/main/scripts/convert_diffusers_to_original_stable_diffusion.py

# Script for converting a HF Diffusers saved pipeline to a Stable Diffusion checkpoint.
# *Only* converts the UNet, VAE, and Text Encoder.
# Does not convert optimizer state or any other thing.

__all__ = [
    # "convert_sd",
    "convert_unet_calibrate_info_sd",
]

import argparse
import os.path as osp
import re
from pathlib import Path

import torch
from safetensors.torch import load_file, save_file

# =================#
# UNet Conversion #
# =================#

unet_conversion_map = [
    # (stable-diffusion, HF Diffusers)
    ("time_embed.0.weight", "time_embedding.linear_1.weight"),
    ("time_embed.0.bias", "time_embedding.linear_1.bias"),
    ("time_embed.2.weight", "time_embedding.linear_2.weight"),
    ("time_embed.2.bias", "time_embedding.linear_2.bias"),
    ("input_blocks.0.0.weight", "conv_in.weight"),
    ("input_blocks.0.0.bias", "conv_in.bias"),
    ("out.0.weight", "conv_norm_out.weight"),
    ("out.0.bias", "conv_norm_out.bias"),
    ("out.2.weight", "conv_out.weight"),
    ("out.2.bias", "conv_out.bias"),
]

unet_conversion_map_resnet = [
    # (stable-diffusion, HF Diffusers)
    ("in_layers.0", "norm1"),
    ("in_layers.2", "conv1"),
    ("out_layers.0", "norm2"),
    ("out_layers.3", "conv2"),
    ("emb_layers.1", "time_emb_proj"),
    ("skip_connection", "conv_shortcut"),
]

unet_conversion_map_layer = []
# hardcoded number of downblocks and resnets/attentions...
# would need smarter logic for other networks.
for i in range(4):
    # loop over downblocks/upblocks

    for j in range(2):
        # loop over resnets/attentions for downblocks
        hf_down_res_prefix = f"down_blocks.{i}.resnets.{j}."
        sd_down_res_prefix = f"input_blocks.{3*i + j + 1}.0."
        unet_conversion_map_layer.append((sd_down_res_prefix, hf_down_res_prefix))

        if i < 3:
            # no attention layers in down_blocks.3
            hf_down_atn_prefix = f"down_blocks.{i}.attentions.{j}."
            sd_down_atn_prefix = f"input_blocks.{3*i + j + 1}.1."
            unet_conversion_map_layer.append((sd_down_atn_prefix, hf_down_atn_prefix))

    for j in range(3):
        # loop over resnets/attentions for upblocks
        hf_up_res_prefix = f"up_blocks.{i}.resnets.{j}."
        sd_up_res_prefix = f"output_blocks.{3*i + j}.0."
        unet_conversion_map_layer.append((sd_up_res_prefix, hf_up_res_prefix))

        if i > 0:
            # no attention layers in up_blocks.0
            hf_up_atn_prefix = f"up_blocks.{i}.attentions.{j}."
            sd_up_atn_prefix = f"output_blocks.{3*i + j}.1."
            unet_conversion_map_layer.append((sd_up_atn_prefix, hf_up_atn_prefix))

    if i < 3:
        # no downsample in down_blocks.3
        hf_downsample_prefix = f"down_blocks.{i}.downsamplers.0.conv."
        sd_downsample_prefix = f"input_blocks.{3*(i+1)}.0.op."
        unet_conversion_map_layer.append((sd_downsample_prefix, hf_downsample_prefix))

        # no upsample in up_blocks.3
        hf_upsample_prefix = f"up_blocks.{i}.upsamplers.0."
        sd_upsample_prefix = f"output_blocks.{3*i + 2}.{1 if i == 0 else 2}."
        unet_conversion_map_layer.append((sd_upsample_prefix, hf_upsample_prefix))

hf_mid_atn_prefix = "mid_block.attentions.0."
sd_mid_atn_prefix = "middle_block.1."
unet_conversion_map_layer.append((sd_mid_atn_prefix, hf_mid_atn_prefix))

for j in range(2):
    hf_mid_res_prefix = f"mid_block.resnets.{j}."
    sd_mid_res_prefix = f"middle_block.{2*j}."
    unet_conversion_map_layer.append((sd_mid_res_prefix, hf_mid_res_prefix))


def convert_unet_state_dict(unet_state_dict):
    # buyer beware: this is a *brittle* function,
    # and correct output requires that all of these pieces interact in
    # the exact order in which I have arranged them.
    mapping = {k: k for k in unet_state_dict.keys()}
    for sd_name, hf_name in unet_conversion_map:
        mapping[hf_name] = sd_name
    for k, v in mapping.items():
        if "resnets" in k:
            for sd_part, hf_part in unet_conversion_map_resnet:
                v = v.replace(hf_part, sd_part)
            mapping[k] = v
    for k, v in mapping.items():
        for sd_part, hf_part in unet_conversion_map_layer:
            v = v.replace(hf_part, sd_part)
        mapping[k] = v
    new_state_dict = {v: unet_state_dict[k] for k, v in mapping.items()}
    return new_state_dict


# ================#
# VAE Conversion #
# ================#

vae_conversion_map = [
    # (stable-diffusion, HF Diffusers)
    ("nin_shortcut", "conv_shortcut"),
    ("norm_out", "conv_norm_out"),
    ("mid.attn_1.", "mid_block.attentions.0."),
]

for i in range(4):
    # down_blocks have two resnets
    for j in range(2):
        hf_down_prefix = f"encoder.down_blocks.{i}.resnets.{j}."
        sd_down_prefix = f"encoder.down.{i}.block.{j}."
        vae_conversion_map.append((sd_down_prefix, hf_down_prefix))

    if i < 3:
        hf_downsample_prefix = f"down_blocks.{i}.downsamplers.0."
        sd_downsample_prefix = f"down.{i}.downsample."
        vae_conversion_map.append((sd_downsample_prefix, hf_downsample_prefix))

        hf_upsample_prefix = f"up_blocks.{i}.upsamplers.0."
        sd_upsample_prefix = f"up.{3-i}.upsample."
        vae_conversion_map.append((sd_upsample_prefix, hf_upsample_prefix))

    # up_blocks have three resnets
    # also, up blocks in hf are numbered in reverse from sd
    for j in range(3):
        hf_up_prefix = f"decoder.up_blocks.{i}.resnets.{j}."
        sd_up_prefix = f"decoder.up.{3-i}.block.{j}."
        vae_conversion_map.append((sd_up_prefix, hf_up_prefix))

# this part accounts for mid blocks in both the encoder and the decoder
for i in range(2):
    hf_mid_res_prefix = f"mid_block.resnets.{i}."
    sd_mid_res_prefix = f"mid.block_{i+1}."
    vae_conversion_map.append((sd_mid_res_prefix, hf_mid_res_prefix))


vae_conversion_map_attn = [
    # (stable-diffusion, HF Diffusers)
    ("norm.", "group_norm."),
    ("q.", "query."),
    ("k.", "key."),
    ("v.", "value."),
    ("proj_out.", "proj_attn."),
]

# This is probably not the most ideal solution, but it does work.
vae_extra_conversion_map = [
    ("to_q", "q"),
    ("to_k", "k"),
    ("to_v", "v"),
    ("to_out.0", "proj_out"),
]


def reshape_weight_for_sd(w):
    # convert HF linear weights to SD conv2d weights
    if not w.ndim == 1:
        return w.reshape(*w.shape, 1, 1)
    else:
        return w


def convert_vae_state_dict(vae_state_dict):
    mapping = {k: k for k in vae_state_dict.keys()}
    for k, v in mapping.items():
        for sd_part, hf_part in vae_conversion_map:
            v = v.replace(hf_part, sd_part)
        mapping[k] = v
    for k, v in mapping.items():
        if "attentions" in k:
            for sd_part, hf_part in vae_conversion_map_attn:
                v = v.replace(hf_part, sd_part)
            mapping[k] = v
    new_state_dict = {v: vae_state_dict[k] for k, v in mapping.items()}
    weights_to_convert = ["q", "k", "v", "proj_out"]
    keys_to_rename = {}
    for k, v in new_state_dict.items():
        for weight_name in weights_to_convert:
            if f"mid.attn_1.{weight_name}.weight" in k:
                print(f"Reshaping {k} for SD format")
                new_state_dict[k] = reshape_weight_for_sd(v)
        for weight_name, real_weight_name in vae_extra_conversion_map:
            if (
                f"mid.attn_1.{weight_name}.weight" in k
                or f"mid.attn_1.{weight_name}.bias" in k
            ):
                keys_to_rename[k] = k.replace(weight_name, real_weight_name)
    for k, v in keys_to_rename.items():
        if k in new_state_dict:
            print(f"Renaming {k} to {v}")
            new_state_dict[v] = reshape_weight_for_sd(new_state_dict[k])
            del new_state_dict[k]
    return new_state_dict


# =========================#
# Text Encoder Conversion #
# =========================#


textenc_conversion_lst = [
    # (stable-diffusion, HF Diffusers)
    ("resblocks.", "text_model.encoder.layers."),
    ("ln_1", "layer_norm1"),
    ("ln_2", "layer_norm2"),
    (".c_fc.", ".fc1."),
    (".c_proj.", ".fc2."),
    (".attn", ".self_attn"),
    ("ln_final.", "transformer.text_model.final_layer_norm."),
    (
        "token_embedding.weight",
        "transformer.text_model.embeddings.token_embedding.weight",
    ),
    (
        "positional_embedding",
        "transformer.text_model.embeddings.position_embedding.weight",
    ),
]
protected = {re.escape(x[1]): x[0] for x in textenc_conversion_lst}
textenc_pattern = re.compile("|".join(protected.keys()))

# Ordering is from https://github.com/pytorch/pytorch/blob/master/test/cpp/api/modules.cpp
code2idx = {"q": 0, "k": 1, "v": 2}


def convert_text_enc_state_dict_v20(text_enc_dict):
    new_state_dict = {}
    capture_qkv_weight = {}
    capture_qkv_bias = {}
    for k, v in text_enc_dict.items():
        if (
            k.endswith(".self_attn.q_proj.weight")
            or k.endswith(".self_attn.k_proj.weight")
            or k.endswith(".self_attn.v_proj.weight")
        ):
            k_pre = k[: -len(".q_proj.weight")]
            k_code = k[-len("q_proj.weight")]
            if k_pre not in capture_qkv_weight:
                capture_qkv_weight[k_pre] = [None, None, None]
            capture_qkv_weight[k_pre][code2idx[k_code]] = v
            continue

        if (
            k.endswith(".self_attn.q_proj.bias")
            or k.endswith(".self_attn.k_proj.bias")
            or k.endswith(".self_attn.v_proj.bias")
        ):
            k_pre = k[: -len(".q_proj.bias")]
            k_code = k[-len("q_proj.bias")]
            if k_pre not in capture_qkv_bias:
                capture_qkv_bias[k_pre] = [None, None, None]
            capture_qkv_bias[k_pre][code2idx[k_code]] = v
            continue

        relabelled_key = textenc_pattern.sub(
            lambda m: protected[re.escape(m.group(0))], k
        )
        new_state_dict[relabelled_key] = v

    for k_pre, tensors in capture_qkv_weight.items():
        if None in tensors:
            raise Exception(
                "CORRUPTED MODEL: one of the q-k-v values for the text encoder was missing"
            )
        relabelled_key = textenc_pattern.sub(
            lambda m: protected[re.escape(m.group(0))], k_pre
        )
        new_state_dict[relabelled_key + ".in_proj_weight"] = torch.cat(tensors)

    for k_pre, tensors in capture_qkv_bias.items():
        if None in tensors:
            raise Exception(
                "CORRUPTED MODEL: one of the q-k-v values for the text encoder was missing"
            )
        relabelled_key = textenc_pattern.sub(
            lambda m: protected[re.escape(m.group(0))], k_pre
        )
        new_state_dict[relabelled_key + ".in_proj_bias"] = torch.cat(tensors)

    return new_state_dict


def convert_text_enc_state_dict(text_enc_dict):
    return text_enc_dict


def convert_unet_calibrate_dict(state_dict) -> str:
    mapping = {k: k for k in state_dict}
    remove_suffix = (
        lambda x: x.removesuffix(".weight").removesuffix(".bias").removesuffix(".")
    )

    for sd_name, hf_name in unet_conversion_map:
        sd_name, hf_name = remove_suffix(sd_name), remove_suffix(hf_name)
        if hf_name in mapping:
            mapping[hf_name] = sd_name
    for k, v in mapping.items():
        if "resnets" in k:
            for sd_part, hf_part in unet_conversion_map_resnet:
                sd_part, hf_part = remove_suffix(sd_part), remove_suffix(hf_part)
                v = v.replace(hf_part, sd_part)
            mapping[k] = v
    for k, v in mapping.items():
        for sd_part, hf_part in unet_conversion_map_layer:
            sd_part, hf_part = remove_suffix(sd_part), remove_suffix(hf_part)
            v = v.replace(hf_part, sd_part)
        mapping[k] = v
    new_state_dict = {v: state_dict[k] for k, v in mapping.items()}
    return new_state_dict


def convert_unet_calibrate_info_sd(calibration_path, dst_path):
    if not Path(calibration_path).exists():
        print(f"File {calibration_path} not found, only convert model")
        return

    calibrate_info = {}
    with open(calibration_path, "r") as f:
        for line in f.readlines():
            line = line.strip()
            items = line.split(" ")
            calibrate_info[items[0]] = " ".join(items[1:])
    dst_info = convert_unet_calibrate_dict(calibrate_info)
    with open(dst_path, "w") as f:
        for name, info in dst_info.items():
            f.write(f"{name} {info}\n")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--model_path",
        default=None,
        type=str,
        required=True,
        help="Path to the model to convert.",
    )
    parser.add_argument(
        "--checkpoint_path",
        default=None,
        type=str,
        required=True,
        help="Path to the output model.",
    )
    parser.add_argument(
        "--half", action="store_true", help="Save weights in half precision."
    )
    parser.add_argument(
        "--use_safetensors",
        action="store_true",
        help="Save weights use safetensors, default is ckpt.",
    )

    args = parser.parse_args()

    assert args.model_path is not None, "Must provide a model path!"

    assert args.checkpoint_path is not None, "Must provide a checkpoint path!"

    # Path for safetensors
    unet_path = osp.join(args.model_path, "unet", "diffusion_pytorch_model.safetensors")
    vae_path = osp.join(args.model_path, "vae", "diffusion_pytorch_model.safetensors")
    text_enc_path = osp.join(args.model_path, "text_encoder", "model.safetensors")

    # Load models from safetensors if it exists, if it doesn't pytorch
    if osp.exists(unet_path):
        unet_state_dict = load_file(unet_path, device="cpu")
    else:
        unet_path = osp.join(args.model_path, "unet", "diffusion_pytorch_model.bin")
        unet_state_dict = torch.load(unet_path, map_location="cpu")

    if osp.exists(vae_path):
        vae_state_dict = load_file(vae_path, device="cpu")
    else:
        vae_path = osp.join(args.model_path, "vae", "diffusion_pytorch_model.bin")
        vae_state_dict = torch.load(vae_path, map_location="cpu")

    if osp.exists(text_enc_path):
        text_enc_dict = load_file(text_enc_path, device="cpu")
    else:
        text_enc_path = osp.join(args.model_path, "text_encoder", "pytorch_model.bin")
        text_enc_dict = torch.load(text_enc_path, map_location="cpu")

    # Convert the UNet model
    unet_state_dict = convert_unet_state_dict(unet_state_dict)
    unet_state_dict = {
        "model.diffusion_model." + k: v for k, v in unet_state_dict.items()
    }

    # Convert the VAE model
    vae_state_dict = convert_vae_state_dict(vae_state_dict)
    vae_state_dict = {"first_stage_model." + k: v for k, v in vae_state_dict.items()}

    # Easiest way to identify v2.0 model seems to be that the text encoder (OpenCLIP) is deeper
    is_v20_model = "text_model.encoder.layers.22.layer_norm2.bias" in text_enc_dict

    if is_v20_model:
        # Need to add the tag 'transformer' in advance so we can knock it out from the final layer-norm
        text_enc_dict = {"transformer." + k: v for k, v in text_enc_dict.items()}
        text_enc_dict = convert_text_enc_state_dict_v20(text_enc_dict)
        text_enc_dict = {
            "cond_stage_model.model." + k: v for k, v in text_enc_dict.items()
        }
    else:
        text_enc_dict = convert_text_enc_state_dict(text_enc_dict)
        text_enc_dict = {
            "cond_stage_model.transformer." + k: v for k, v in text_enc_dict.items()
        }

    # Put together new checkpoint
    state_dict = {**unet_state_dict, **vae_state_dict, **text_enc_dict}
    if args.half:
        state_dict = {k: v.half() for k, v in state_dict.items()}

    if args.use_safetensors:
        save_file(state_dict, args.checkpoint_path)
    else:
        state_dict = {"state_dict": state_dict}
        torch.save(state_dict, args.checkpoint_path)

    calibrate_info_save_path = (
        Path(args.checkpoint_path).parent
        / f"{Path(args.checkpoint_path).stem}_sd_calibrate_info.txt"
    )
    convert_unet_calibrate_info_sd(
        args.model_path + "/calibrate_info.txt", calibrate_info_save_path
    )

# def get_unet_state_dict(model_path):
#     unet_path = osp.join(model_path, "unet", "diffusion_pytorch_model.safetensors")
#     if osp.exists(unet_path):
#         unet_state_dict = load_file(unet_path, device="cpu")
#     else:
#         unet_path = osp.join(model_path, "unet", "diffusion_pytorch_model.bin")
#         unet_state_dict = torch.load(unet_path, map_location="cpu")

#     # Convert the UNet model
#     unet_state_dict = convert_unet_state_dict(unet_state_dict)
#     unet_state_dict = {
#         "model.diffusion_model." + k: v for k, v in unet_state_dict.items()
#     }
#     return unet_state_dict


# def get_vae_state_dict(model_path):
#     vae_path = osp.join(model_path, "vae", "diffusion_pytorch_model.safetensors")
#     if osp.exists(vae_path):
#         vae_state_dict = load_file(vae_path, device="cpu")
#     else:
#         vae_path = osp.join(model_path, "vae", "diffusion_pytorch_model.bin")
#         vae_state_dict = torch.load(vae_path, map_location="cpu")

#     # Convert the VAE model
#     vae_state_dict = convert_vae_state_dict(vae_state_dict)
#     vae_state_dict = {"first_stage_model." + k: v for k, v in vae_state_dict.items()}
#     return vae_state_dict


# def get_text_enc_state_dict(model_path):
#     text_enc_path = osp.join(model_path, "text_encoder", "model.safetensors")
#     if osp.exists(text_enc_path):
#         text_enc_dict = load_file(text_enc_path, device="cpu")
#     else:
#         text_enc_path = osp.join(model_path, "text_encoder", "pytorch_model.bin")
#         text_enc_dict = torch.load(text_enc_path, map_location="cpu")

#     # Easiest way to identify v2.0 model seems to be that the text encoder (OpenCLIP) is deeper
#     is_v20_model = "text_model.encoder.layers.22.layer_norm2.bias" in text_enc_dict

#     if is_v20_model:
#         # Need to add the tag 'transformer' in advance so we can knock it out from the final layer-norm
#         text_enc_dict = {"transformer." + k: v for k, v in text_enc_dict.items()}
#         text_enc_dict = convert_text_enc_state_dict_v20(text_enc_dict)
#         text_enc_dict = {
#             "cond_stage_model.model." + k: v for k, v in text_enc_dict.items()
#         }
#     else:
#         text_enc_dict = convert_text_enc_state_dict(text_enc_dict)
#         text_enc_dict = {
#             "cond_stage_model.transformer." + k: v for k, v in text_enc_dict.items()
#         }

#     return text_enc_dict


# def convert_sd(model_path, checkpoint_path, *, half=False, use_safetensors=True):
#     # Put together new checkpoint
#     unet_state_dict = get_unet_state_dict(model_path)
#     vae_state_dict = get_vae_state_dict(model_path)
#     text_enc_dict = get_text_enc_state_dict(model_path)
#     state_dict = {**unet_state_dict, **vae_state_dict, **text_enc_dict}
#     if half:
#         state_dict = {k: v.half() for k, v in state_dict.items()}

#     if use_safetensors:
#         save_file(state_dict, checkpoint_path)
#     else:
#         state_dict = {"state_dict": state_dict}
#         torch.save(state_dict, checkpoint_path)



import os
from functools import singledispatchmethod

from comfy.controlnet import ControlLora, ControlNet
from comfy.model_patcher import ModelPatcher

from comfy.sd import VAE
from onediff.infer_compiler.backends.oneflow import (
    OneflowDeployableModule as DeployableModule,
)

from ..booster_interface import BoosterExecutor


class PatchBoosterExecutor(BoosterExecutor):
    @singledispatchmethod
    def execute(self, model, ckpt_name=None):
        return model

    def _set_batch_size_patch(self, diff_model: DeployableModule, latent_image):
        batch_size = latent_image["samples"].shape[0]
        if isinstance(diff_model, DeployableModule):
            file_path = diff_model.get_graph_file()
            if file_path is None:
                return diff_model

            file_dir = os.path.dirname(file_path)
            file_name = os.path.basename(file_path)
            names = file_name.split("_")
            key, is_replace = "bs=", False
            for i, name in enumerate(names):
                if key in name:
                    names[i] = f"{key}{batch_size}"
                    is_replace = True
            if not is_replace:
                names = [f"{key}{batch_size}"] + names

            new_file_name = "_".join(names)
            new_file_path = os.path.join(file_dir, new_file_name)

            diff_model.set_graph_file(new_file_path)
        else:
            print(f"Warning: model is not a {DeployableModule}")
        return diff_model

    @execute.register(ModelPatcher)
    def _(self, model, **kwargs):
        latent_image = kwargs.get("latent_image", None)
        if latent_image:
            diff_model = model.model.diffusion_model
            self._set_batch_size_patch(diff_model, latent_image)
        return model



import os
import traceback

COMFYUI_ROOT = os.getenv("COMFYUI_ROOT")
from onediff.utils.import_utils import DynamicModuleLoader

from ...sd_hijack_utils import Hijacker

__all__ = ["comfyui_instantid"]

pkg_name = "ComfyUI_InstantID"
pkg_root = os.path.join(COMFYUI_ROOT, "custom_nodes", pkg_name)
is_load_comfyui_instantid_pkg = True
try:
    if os.path.exists(pkg_root):
        comfyui_instantid = DynamicModuleLoader.from_path(pkg_root)
    else:
        is_load_comfyui_instantid_pkg = False
except Exception as e:
    print(traceback.format_exc())
    print(f"Warning: Failed to load {pkg_root} due to {e}")
    is_load_comfyui_instantid_pkg = False
comfyui_instantid_hijacker = Hijacker()



import argparse

import matplotlib as mpl
import matplotlib.pyplot as plt
import yaml


def parse_args():
    parser = argparse.ArgumentParser(description="Plot optimization scores")
    parser.add_argument(
        "--config", type=str, default="config.yaml", help="Path to the config file"
    )
    return parser.parse_args()


# Load settings from the YAML file
args = parse_args()
with open(args.config, "r") as file:
    config = yaml.safe_load(file)

categories = config["categories"]
optimization_scores = config["optimization_scores"]

# Use default scores on bars if configured to do so
if config.get("use_default_scores_on_bars", False):
    optimization_scores_on_bars = config.get("optimization_scores", [0.3, 0.3, 0.35])
else:
    optimization_scores_on_bars = config.get("custom_scores_on_bars", [])


# Set up a colormap for gradient color
colormap_range = config.get("colormap_range", [0.2, 0.5, 0.8])
# colors = plt.cm.get_cmap(config['colormap'])(colormap_range)
colors = mpl.colormaps.get_cmap(config["colormap"])(colormap_range)


# Set figure size and resolution
fig, ax = plt.subplots(
    figsize=(
        config["figure"]["figsize"]["width"],
        config["figure"]["figsize"]["height"],
    ),
    dpi=config["figure"]["dpi"],
)

# Customize bar width and use gradient colors
bars = ax.bar(
    categories,
    optimization_scores,
    width=config["bar_width"],
    color=colors,
    edgecolor="None",
)

# Add Optimization Scores on top of the bars
if config.get("display_scores_on_bars", False):
    for bar, score in zip(bars, optimization_scores_on_bars):
        ax.text(
            bar.get_x() + bar.get_width() / 2,
            bar.get_height() + 0.01,
            config.get("scores_format_on_bars", "{score:.2f}").format(score=score),
            ha="center",
            va="bottom",
            fontsize=config["label_fontsize"],
        )
        # ax.text(bar.get_x() + bar.get_width() / 2, bar.get_height() + 0.01, f'{score:.2f}', ha='center', va='bottom', fontsize=config['label_fontsize'])

# Adjust font size for ticks
ax.set_xticks(ax.get_xticks())
ax.set_xticklabels(
    categories,
    fontsize=config["label_fontsize"],
    rotation=config["rotation_angle"],
    ha="right",
)
ax.tick_params(
    axis="x", which="both", direction="in"
)  # Set x-axis ticks direction to 'in'

ax.set_yticks(ax.get_yticks())
ax.set_yticklabels(
    [f"{tick:.2f}" for tick in ax.get_yticks()], fontsize=config["label_fontsize"]
)
ax.tick_params(
    axis="y", which="both", direction="in"
)  # Set y-axis ticks direction to 'in'

# Remove right and top spines
ax.spines["right"].set_visible(False)
ax.spines["top"].set_visible(False)

# Set the color of dashed lines for right and top frame
ax.spines["right"].set_linestyle(config["line_style"])
ax.spines["top"].set_linestyle(config["line_style"])
ax.spines["right"].set_edgecolor(config["line_color"])
ax.spines["top"].set_edgecolor(config["line_color"])

# Add dashed grid lines for both x-axis and y-axis ticks
ax.grid(
    axis="both",
    linestyle=config["line_style"],
    linewidth=config["line_width"],
    color=config["line_color"],
    alpha=config["line_alpha"],
)

# Set labels and title
ax.set_xlabel(config["x_label"], fontsize=config["label_fontsize"])
ax.set_ylabel(config["y_label"], fontsize=config["label_fontsize"])
ax.set_title(config["plot_title"], fontsize=config["label_fontsize"])

# Save the plot with higher quality
plt.savefig(config["output_filename"], dpi=config["figure"]["dpi"], bbox_inches="tight")

# Show the plot
plt.show()



import oneflow as torch
from onediff.infer_compiler.backends.oneflow.transform import transform_mgr

transformed_comfy = transform_mgr.transform_package("comfy")
proxy_ops = transformed_comfy.ops

if hasattr(proxy_ops, "disable_weight_init"):
    proxy_linear = proxy_ops.disable_weight_init.Linear
else:
    proxy_linear = proxy_ops.Linear


class Linear(proxy_linear):
    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = torch.nn.Parameter(
            torch.empty((out_features, in_features), **factory_kwargs)
        )
        if bias:
            self.bias = torch.nn.Parameter(torch.empty(out_features, **factory_kwargs))
        else:
            self.register_parameter("bias", None)

    def forward(self, input):
        import os

        if (
            self.bias is not None
            and os.getenv("ONEFLOW_KERNEL_ENABLE_FUSED_LINEAR") == "1"
        ):
            return torch._C.fused_matmul_bias(input, self.weight, self.bias)
        else:
            return torch.nn.functional.linear(input, self.weight, self.bias)



from modules import shared
from modules.sd_vae_approx import model as get_vae_model, sd_vae_approx_models

from onediff.infer_compiler import compile

from compile.utils import get_onediff_backend

__all__ = ["VaeCompileCtx"]

compiled_models = {}


class VaeCompileCtx(object):
    def __init__(self, backend=None, options=None):
        self._options = options
        # https://github.com/AUTOMATIC1111/stable-diffusion-webui/blob/75336dfc84cae280036bc52a6805eb10d9ae30ba/modules/sd_vae_approx.py#L43
        self._model_name = (
            "vaeapprox-sdxl.pt"
            if getattr(shared.sd_model, "is_sdxl", False)
            else "model.pt"
        )
        self._original_model = get_vae_model()
        self.backend = backend

    def __enter__(self):
        if self._original_model is None:
            return
        global compiled_models
        model = compiled_models.get(self._model_name)
        backend = str(self.backend or get_onediff_backend())
        if model is None:
            model = compile(
                self._original_model, backend=backend, options=self._options
            )
            compiled_models[self._model_name] = model
        sd_vae_approx_models[self._model_name] = model

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._original_model is not None:
            sd_vae_approx_models[self._model_name] = self._original_model
        return False



"""
Before running this script, you need to start the Selenium and ComfyUI services.
You can start their containers using Docker Compose.

Please set the following environment variables (whose values are for reference only):

export ACR_ORG=registry.cn-beijing.aliyuncs.com/oneflow
export MATRIX_IMAGE=onediff-pro:cu121
export COMFYUI_SRC_DIR=ComfyUI
export SELENIUM_IMAGE=standalone-chrome:119.0-chromedriver-119.0-grid-4.15.0-20231129
export SELENIUM_CONTAINER_NAME=selenium-test
export COMFYUI_PORT=8855
export CONTAINER_NAME=onediff-test
export SDXL_BASE=/share_nfs/hf_models/sd_xl_base_1.0.safetensors
export UNET_INT8=/share_nfs/hf_models/unet_int8
export CONTROL_LORA_OPENPOSEXL2_RANK256=/share_nfs/hf_models/controlnet/control-lora-openposeXL2-rank256.safetensors

And then:

**Note**:
  It is advisable to execute the following commands in the 'diffusers' directory
unless you are fully aware of the implications of executing them in a different
directory.

git clone https://github.com/comfyanonymous/ComfyUI.git

docker compose -f tests/comfy-docker-compose.yml up -d
docker exec $CONTAINER_NAME python3 -m pip config set global.index-url https://pypi.tuna.tsinghua.edu.cn/simple
docker exec $CONTAINER_NAME python3 -m pip install -r tests/comfyui/requirements.txt --user
docker exec $CONTAINER_NAME python3 -m pip install -r /app/ComfyUI/requirements.txt --user
docker exec -it $CONTAINER_NAME python3 /app/ComfyUI/main.py --cuda-device 0

Run the test script:

python tests/comfyui/test_by_ui.py --comfy_port 8188 --workflow tests/comfyui/workflows/sdxl-unet-speedup-graph-saver.json

If you need to shutdown the test containers, run:

docker compose -f tests/comfy-docker-compose.yml down
"""
import argparse
import os
import time

from PIL import Image
from selenium import webdriver
from selenium.common.exceptions import TimeoutException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait


def parse_args():
    parser = argparse.ArgumentParser(description="Test ComfyUI workflow by Selenium.")
    parser.add_argument(
        "-w",
        "--workflow",
        type=str,
        required=True,
        help="Workflow file",
    )
    parser.add_argument(
        "-t",
        "--timeout",
        type=int,
        default="200",
    )
    parser.add_argument(
        "--host",
        type=str,
        default="127.0.0.1",
        help="The selenium service host",
    )
    parser.add_argument(
        "--port",
        type=str,
        default="4444",
        help="The selenium service port",
    )
    parser.add_argument(
        "--comfy_port",
        type=str,
        default="8188",
        help="The ComfyUI service port",
    )
    args = parser.parse_args()
    return args


args = parse_args()


def extract_metadata_from_png(png_file_path):
    img = Image.open(png_file_path)
    metadata = img.info
    img.close()
    return metadata


def read_workflow_json(filename) -> str:
    _, extension = os.path.splitext(filename)
    if extension.endswith("json"):
        with open(filename, "r") as f:
            c = f.read()
            return c
    elif extension.endswith("png"):
        metadata = extract_metadata_from_png(filename)
        return metadata["workflow"]
    else:
        raise NotImplementedError("Only json or png workflow file supported yet")


def load_workflow_graph(driver, workflow: str):
    driver.execute_script(f"window.app.loadGraphData({workflow})")


def click_queue_prompt_button(driver):
    wait = WebDriverWait(driver, 1)
    queue_button = wait.until(EC.presence_of_element_located((By.ID, "queue-button")))
    queue_button.click()


def clear_curernt_workflow(driver):
    driver.execute_script("window.app.graph.clear()")


def wait_until_queue_finished(driver, timeout=100):
    time.sleep(0.3)
    wait = WebDriverWait(driver, timeout)
    element = wait.until(
        EC.presence_of_element_located(
            (By.XPATH, f'//*[contains(text(), "Queue size: 0")]')
        )
    )


def wait_until_app_ready(driver):
    WebDriverWait(driver, 5).until(
        EC.presence_of_element_located((By.ID, "comfy-clear-button"))
    )
    WebDriverWait(driver, 5).until(
        EC.presence_of_element_located((By.ID, "graph-canvas"))
    )


def launch_and_wait(driver, *, timeout=100):
    click_queue_prompt_button(driver)
    wait_until_queue_finished(driver, timeout)


def check_graph_node_types(driver):
    try:
        driver.execute_script("window.graph.checkNodeTypes()")
    except Exception as e:
        raise Exception("Error: Workflow nodes checking failed, likely missing nodes")


def check_error_occurs(driver):
    elements = driver.find_elements(By.CLASS_NAME, "comfy-modal-content")

    desired_element = None
    for element in elements:
        element_text = element.text
        if "Error occurred when" in element_text:
            print(element.text)
            raise Exception(f"{element.text}")


def launch_prompt(driver):
    try:
        print(f"connect to ComfyUI: 127.0.0.1:{args.comfy_port}...")
        driver.get(f"http://127.0.0.1:{args.comfy_port}")
        print(f"ComfyUI connected")
        time.sleep(0.1)
        start_time = time.time()

        wait_until_app_ready(driver)

        print("clear the workflow...")
        clear_curernt_workflow(driver)
        print("workflow cleard")

        print("load the target workflow...")
        load_workflow_graph(driver, read_workflow_json(args.workflow))
        print(f"{args.workflow} loaded")

        print("check the nodes type of workflow...")
        check_graph_node_types(driver)
        print(f"{args.workflow} workflow checked")

        print(f"launch the queue prompt (timeout: {args.timeout}s) ...")
        launch_and_wait(driver, timeout=args.timeout)

        duration = time.time() - start_time
        print(f"{args.workflow} has finished, time elapsed: {duration:.1f}")

        if duration < 2:
            raise ValueError(
                "Execution duration is too short, possible error in workflow execution"
            )

        print(f"check if error occurs...")
        check_error_occurs(driver)
        print(f"no error occurs when executing workflow")
    except TimeoutException:
        print("Time out")
        exit(1)
    except Exception as e:
        print(type(e))
        print(e)
        print("exit with error: 1")
        exit(1)
    finally:
        driver.quit()


if __name__ == "__main__":
    options = webdriver.ChromeOptions()
    print(f"connect remote Selenium: http://{args.host}:{args.port}/wd/hub ...")
    driver = webdriver.Remote(
        command_executor=f"http://{args.host}:{args.port}/wd/hub", options=options
    )
    print(f"Selenium connected")
    launch_prompt(driver)



import dataclasses
from typing import Union

import torch
from ldm.modules.diffusionmodules.openaimodel import UNetModel as LdmUNetModel
from modules import sd_models_types, shared

from onediff.infer_compiler import DeployableModule
from sgm.modules.diffusionmodules.openaimodel import UNetModel as SgmUNetModel

from .backend import OneDiffBackend


def disable_unet_checkpointing(
    unet_model: Union[LdmUNetModel, SgmUNetModel]
) -> Union[LdmUNetModel, SgmUNetModel]:
    from ldm.modules.attention import BasicTransformerBlock as LdmBasicTransformerBlock
    from ldm.modules.diffusionmodules.openaimodel import ResBlock as LdmResBlock
    from sgm.modules.attention import BasicTransformerBlock as SgmBasicTransformerBlock
    from sgm.modules.diffusionmodules.openaimodel import ResBlock as SgmResBlock

    for module in unet_model.modules():
        if isinstance(module, (LdmBasicTransformerBlock, SgmBasicTransformerBlock)):
            module.checkpoint = False
        if isinstance(module, (LdmResBlock, SgmResBlock)):
            module.use_checkpoint = False
    return unet_model


def get_onediff_backend() -> OneDiffBackend:
    return OneDiffBackend(shared.opts.onediff_compiler_backend)


def is_oneflow_backend(backend: Union[OneDiffBackend, None] = None) -> bool:
    return (backend or get_onediff_backend()) == OneDiffBackend.ONEFLOW


def is_nexfort_backend(backend: Union[OneDiffBackend, None] = None) -> bool:
    return (backend or get_onediff_backend()) == OneDiffBackend.NEXFORT


def init_backend(backend: Union[OneDiffBackend, None] = None):
    backend = backend or get_onediff_backend()
    if is_oneflow_backend(backend):
        from .oneflow.utils import init_oneflow_backend

        init_oneflow_backend()
    elif is_nexfort_backend(backend):
        from .nexfort.utils import init_nexfort_backend

        init_nexfort_backend()
    else:
        raise NotImplementedError(f"invalid backend {backend}")


@dataclasses.dataclass
class OneDiffCompiledGraph:
    name: str = None
    filename: str = None
    sha: str = None
    eager_module: torch.nn.Module = None
    graph_module: DeployableModule = None
    quantized: bool = False

    def __init__(
        self,
        sd_model: sd_models_types.WebuiSdModel = None,
        unet_model=None,
        graph_module: DeployableModule = None,
        quantized=False,
    ):
        if sd_model is None:
            return
        self.name = sd_model.sd_checkpoint_info.name
        self.filename = sd_model.sd_checkpoint_info.filename
        self.sha = sd_model.sd_model_hash
        self.eager_module = unet_model or sd_model.model.diffusion_model
        self.graph_module = graph_module
        self.quantized = quantized



import torch
from compile.utils import is_oneflow_backend

from onediff.infer_compiler import DeployableModule


class HijackLoraActivate:
    def __init__(self):
        from modules import extra_networks

        if "lora" in extra_networks.extra_network_registry:
            cls_extra_network_lora = type(extra_networks.extra_network_registry["lora"])
        else:
            cls_extra_network_lora = None
        self.lora_class = cls_extra_network_lora

    def __enter__(self):
        if self.lora_class is None:
            return
        self.orig_func = self.lora_class.activate
        self.lora_class.activate = hijacked_activate(self.lora_class.activate)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.lora_class is None:
            return
        self.lora_class.activate = self.orig_func
        self.lora_class = None
        self.orig_func = None


def hijacked_activate(activate_func):
    import networks

    if hasattr(activate_func, "_onediff_hijacked"):
        return activate_func

    def activate(self, p, params_list):
        activate_func(self, p, params_list)
        if isinstance(p.sd_model.model.diffusion_model, DeployableModule):
            onediff_sd_model: DeployableModule = p.sd_model.model.diffusion_model
            for name, sub_module in onediff_sd_model.named_modules():
                if not isinstance(
                    sub_module,
                    (
                        torch.nn.Linear,
                        torch.nn.Conv2d,
                        torch.nn.GroupNorm,
                        torch.nn.LayerNorm,
                    ),
                ):
                    continue
                networks.network_apply_weights(sub_module)
                if is_oneflow_backend() and isinstance(sub_module, torch.nn.Conv2d):
                    # TODO(WangYi): refine here
                    from onediff.infer_compiler.backends.oneflow.param_utils import (
                        update_graph_related_tensor,
                    )

                    try:
                        update_graph_related_tensor(sub_module)
                    except:
                        pass

    activate._onediff_hijacked = True
    return activate



from onediff.infer_compiler.backends.oneflow.transform import register
from onediff.utils.import_utils import is_oneflow_available
from onediff_utils import singleton_decorator


@singleton_decorator
def init_oneflow_backend():
    if not is_oneflow_available():
        raise RuntimeError(
            "Backend oneflow for OneDiff is invalid, please make sure you have installed OneFlow"
        )

    from .mock import ldm, sgm, vae

    register(package_names=["ldm"], torch2oflow_class_map=ldm.torch2oflow_class_map)
    register(package_names=["sgm"], torch2oflow_class_map=sgm.torch2oflow_class_map)
    register(package_names=["modules"], torch2oflow_class_map=vae.torch2oflow_class_map)



MODEL = "stabilityai/stable-video-diffusion-img2vid-xt"
# SVD 1.1: stabilityai/stable-video-diffusion-img2vid-xt-1-1 is also available.
VARIANT = None
CUSTOM_PIPELINE = None
SCHEDULER = None
LORA = None
CONTROLNET = None
STEPS = 25
SEED = None
WARMUPS = 1
BATCH = 1
ALTER_HEIGHT = None
ALTER_WIDTH = None
# The official recommended parameters for SVD 1.1 are:
# Resolution: 1024x576,
# Frame count: 25 frames,
# FPS: 6,
# Motion Bucket Id: 127.
HEIGHT = 576
WIDTH = 1024
FRAMES = 25
FPS = 7
MOTION_BUCKET_ID = 127
DECODE_CHUNK_SIZE = 5
INPUT_IMAGE = "https://huggingface.co/datasets/huggingface/documentation-images/resolve/main/diffusers/svd/rocket.png?download=true"
CONTROL_IMAGE = None
OUTPUT_VIDEO = None
EXTRA_CALL_KWARGS = None
ATTENTION_FP16_SCORE_ACCUM_MAX_M = 0
CACHE_INTERVAL = 3
CACHE_BRANCH = 0

import argparse
import importlib
import inspect
import json
import os
import random
import time

from PIL import Image, ImageDraw

import oneflow as flow  # usort: skip
import torch
from diffusers.utils import export_to_video, load_image
from onediffx import compile_pipe, OneflowCompileOptions


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", type=str, default=MODEL)
    parser.add_argument("--save_graph", action="store_true")
    parser.add_argument("--load_graph", action="store_true")
    parser.add_argument("--variant", type=str, default=VARIANT)
    parser.add_argument("--custom-pipeline", type=str, default=CUSTOM_PIPELINE)
    parser.add_argument("--scheduler", type=str, default=SCHEDULER)
    parser.add_argument("--lora", type=str, default=LORA)
    parser.add_argument("--controlnet", type=str, default=CONTROLNET)
    parser.add_argument("--steps", type=int, default=STEPS)
    parser.add_argument("--seed", type=int, default=SEED)
    parser.add_argument("--warmups", type=int, default=WARMUPS)
    parser.add_argument("--frames", type=int, default=FRAMES)
    parser.add_argument("--batch", type=int, default=BATCH)
    parser.add_argument("--height", type=int, default=HEIGHT)
    parser.add_argument("--width", type=int, default=WIDTH)
    parser.add_argument("--fps", type=int, default=FPS)
    parser.add_argument("--motion_bucket_id", type=int, default=MOTION_BUCKET_ID)
    parser.add_argument("--decode-chunk-size", type=int, default=DECODE_CHUNK_SIZE)
    parser.add_argument("--cache_interval", type=int, default=CACHE_INTERVAL)
    parser.add_argument("--cache_branch", type=int, default=CACHE_BRANCH)
    parser.add_argument("--extra-call-kwargs", type=str, default=EXTRA_CALL_KWARGS)
    parser.add_argument("--deepcache", action="store_true")
    parser.add_argument("--input-image", type=str, default=INPUT_IMAGE)
    parser.add_argument("--control-image", type=str, default=CONTROL_IMAGE)
    parser.add_argument("--output-video", type=str, default=OUTPUT_VIDEO)
    parser.add_argument(
        "--compiler",
        type=str,
        default="oneflow",
        choices=["none", "oneflow", "compile"],
    )
    parser.add_argument(
        "--attention-fp16-score-accum-max-m",
        type=int,
        default=ATTENTION_FP16_SCORE_ACCUM_MAX_M,
    )
    parser.add_argument(
        "--alter-height",
        type=int,
        default=ALTER_HEIGHT,
    )
    parser.add_argument(
        "--alter-width",
        type=int,
        default=ALTER_WIDTH,
    )
    return parser.parse_args()


def load_pipe(
    pipeline_cls,
    model_name,
    variant=None,
    custom_pipeline=None,
    scheduler=None,
    lora=None,
    controlnet=None,
):
    extra_kwargs = {}
    if custom_pipeline is not None:
        extra_kwargs["custom_pipeline"] = custom_pipeline
    if variant is not None:
        extra_kwargs["variant"] = variant
    if controlnet is not None:
        from diffusers import ControlNetModel

        controlnet = ControlNetModel.from_pretrained(
            controlnet,
            torch_dtype=torch.float16,
        )
        extra_kwargs["controlnet"] = controlnet
    if os.path.exists(os.path.join(model_name, "calibrate_info.txt")):
        from onediff.quantization import QuantPipeline

        pipe = QuantPipeline.from_quantized(
            pipeline_cls, model_name, torch_dtype=torch.float16, **extra_kwargs
        )
    else:
        pipe = pipeline_cls.from_pretrained(
            model_name, torch_dtype=torch.float16, **extra_kwargs
        )
    if scheduler is not None:
        scheduler_cls = getattr(importlib.import_module("diffusers"), scheduler)
        pipe.scheduler = scheduler_cls.from_config(pipe.scheduler.config)
    if lora is not None:
        pipe.load_lora_weights(lora)
        pipe.fuse_lora()
    pipe.safety_checker = None
    pipe.to(torch.device("cuda"))
    return pipe


class IterationProfiler:
    def __init__(self):
        self.begin = None
        self.end = None
        self.num_iterations = 0

    def get_iter_per_sec(self):
        if self.begin is None or self.end is None:
            return None
        self.end.synchronize()
        dur = self.begin.elapsed_time(self.end)
        return self.num_iterations / dur * 1000.0

    def callback_on_step_end(self, pipe, i, t, callback_kwargs={}):
        if self.begin is None:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.begin = event
        else:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.end = event
            self.num_iterations += 1
        return callback_kwargs


def main():
    args = parse_args()
    if args.deepcache:
        from onediffx.deep_cache import StableVideoDiffusionPipeline
    else:
        from diffusers import StableVideoDiffusionPipeline

    pipe = load_pipe(
        StableVideoDiffusionPipeline,
        args.model,
        variant=args.variant,
        custom_pipeline=args.custom_pipeline,
        scheduler=args.scheduler,
        lora=args.lora,
        controlnet=args.controlnet,
    )

    height = args.height or pipe.unet.config.sample_size * pipe.vae_scale_factor
    width = args.width or pipe.unet.config.sample_size * pipe.vae_scale_factor

    if args.compiler == "none":
        pass
    elif args.compiler == "oneflow":
        # The absolute element values of K in the attention layer of SVD is too large.
        # The unfused attention (without SDPA) and MHA with half accumulation would both overflow.
        # But disabling all half accumulations in MHA would slow down the inference,
        # especially for 40xx series cards.
        # So here by partially disabling the half accumulation in MHA partially,
        # we can get a good balance.
        compile_options = OneflowCompileOptions()
        compile_options.attention_allow_half_precision_score_accumulation_max_m = (
            args.attention_fp16_score_accum_max_m
        )
        pipe = compile_pipe(pipe, options=compile_options)
    elif args.compiler == "compile":
        pipe.unet = torch.compile(pipe.unet)
        if hasattr(pipe, "controlnet"):
            pipe.controlnet = torch.compile(pipe.controlnet)
        pipe.vae = torch.compile(pipe.vae)
    else:
        raise ValueError(f"Unknown compiler: {args.compiler}")

    resolutions = [[height, width]]
    if args.alter_height is not None:
        # Test dynamic shape.
        assert args.alter_width is not None
        resolutions.append([args.alter_height, args.alter_width])
    for height, width in resolutions:
        input_image = load_image(args.input_image)
        input_image.resize((width, height), Image.LANCZOS)

        if args.control_image is None:
            if args.controlnet is None:
                control_image = None
            else:
                control_image = Image.new("RGB", (width, height))
                draw = ImageDraw.Draw(control_image)
                draw.ellipse(
                    (
                        width // 4,
                        height // 4,
                        width // 4 * 3,
                        height // 4 * 3,
                    ),
                    fill=(255, 255, 255),
                )
                del draw
        else:
            control_image = load_image(args.control_image)
            control_image = control_image.resize((args.width, height), Image.LANCZOS)

        def get_kwarg_inputs():
            kwarg_inputs = dict(
                image=input_image,
                height=height,
                width=width,
                num_inference_steps=args.steps,
                num_videos_per_prompt=args.batch,
                num_frames=args.frames,
                fps=args.fps,
                motion_bucket_id=args.motion_bucket_id,
                decode_chunk_size=args.decode_chunk_size,
                generator=None
                if args.seed is None
                else torch.Generator().manual_seed(args.seed),
                **(
                    dict()
                    if args.extra_call_kwargs is None
                    else json.loads(args.extra_call_kwargs)
                ),
            )
            if control_image is not None:
                kwarg_inputs["control_image"] = control_image
            if args.deepcache:
                kwarg_inputs["cache_interval"] = args.cache_interval
                kwarg_inputs["cache_branch"] = args.cache_branch
            return kwarg_inputs

        if args.warmups > 0:
            if args.load_graph:
                print("Loading graphs to avoid compilation...")
                start_t = time.time()
                pipe.unet.load_graph("base_unet_compiled", run_warmup=True)
                pipe.vae.decoder.load_graph("base_vae_compiled", run_warmup=True)
                end_t = time.time()
                print(f"Loading graph elapsed: {end_t - start_t} s")
                print("Begin warmup")
                for _ in range(args.warmups):
                    pipe(**get_kwarg_inputs())
                print("End warmup")
            else:
                print("Begin warmup")
                for _ in range(args.warmups):
                    pipe(**get_kwarg_inputs())
                print("End warmup")

        kwarg_inputs = get_kwarg_inputs()
        iter_profiler = IterationProfiler()
        if "callback_on_step_end" in inspect.signature(pipe).parameters:
            kwarg_inputs["callback_on_step_end"] = iter_profiler.callback_on_step_end
        elif "callback" in inspect.signature(pipe).parameters:
            kwarg_inputs["callback"] = iter_profiler.callback_on_step_end
        begin = time.time()
        output_frames = pipe(**kwarg_inputs).frames
        end = time.time()

        print(f"Inference time: {end - begin:.3f}s")
        iter_per_sec = iter_profiler.get_iter_per_sec()
        if iter_per_sec is not None:
            print(f"Iterations per second: {iter_per_sec:.3f}")
        cuda_mem_after_used = flow._oneflow_internal.GetCUDAMemoryUsed()
        host_mem_after_used = flow._oneflow_internal.GetCPUMemoryUsed()
        print(f"CUDA Mem after: {cuda_mem_after_used / 1024:.3f}GiB")
        print(f"Host Mem after: {host_mem_after_used / 1024:.3f}GiB")

        if args.output_video is not None:
            export_to_video(output_frames[0], args.output_video, fps=args.fps)
        else:
            print("Please set `--output-video` to save the output video")

    if args.save_graph:
        print("Saving graphs...")
        start_t = time.time()
        pipe.unet.save_graph("base_unet_compiled")
        pipe.vae.decoder.save_graph("base_vae_compiled")
        end_t = time.time()
        print(f"save graphs elapsed: {end_t - start_t} s")


if __name__ == "__main__":
    main()



MODEL = "stabilityai/stable-diffusion-xl-base-1.0"
REPO = "ByteDance/SDXL-Lightning"
CPKT = "sdxl_lightning_4step_unet.safetensors"
VARIANT = None
CUSTOM_PIPELINE = None
CONTROLNET = None
PROMPT = "A girl smiling"
NEGATIVE_PROMPT = None
SEED = None
WARMUPS = 3
BATCH = 1
HEIGHT = 1024
WIDTH = 1024
OUTPUT_IMAGE = None
EXTRA_CALL_KWARGS = None

import argparse
import importlib
import inspect
import json
import os
import time

import torch
from diffusers.utils import load_image
from PIL import Image, ImageDraw

import oneflow as flow  # usort: skip
from huggingface_hub import hf_hub_download
from onediffx import compile_pipe
from safetensors.torch import load_file


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", type=str, default=MODEL)
    parser.add_argument("--repo", type=str, default=REPO)
    parser.add_argument("--cpkt", type=str, default=CPKT)
    parser.add_argument("--variant", type=str, default=VARIANT)
    parser.add_argument("--custom-pipeline", type=str, default=CUSTOM_PIPELINE)
    parser.add_argument("--controlnet", type=str, default=CONTROLNET)
    parser.add_argument("--prompt", type=str, default=PROMPT)
    parser.add_argument("--negative-prompt", type=str, default=NEGATIVE_PROMPT)
    parser.add_argument("--seed", type=int, default=SEED)
    parser.add_argument("--warmups", type=int, default=WARMUPS)
    parser.add_argument("--batch", type=int, default=BATCH)
    parser.add_argument("--height", type=int, default=HEIGHT)
    parser.add_argument("--width", type=int, default=WIDTH)
    parser.add_argument("--extra-call-kwargs", type=str, default=EXTRA_CALL_KWARGS)
    parser.add_argument("--output-image", type=str, default=OUTPUT_IMAGE)
    parser.add_argument(
        "--compiler",
        type=str,
        default="oneflow",
        choices=["none", "oneflow", "compile", "compile-max-autotune"],
    )
    return parser.parse_args()


def load_and_compile_pipe(
    model_name,
    repo_name,
    cpkt_name,
    compile_type,
    variant=None,
    custom_pipeline=None,
    controlnet=None,
):
    from diffusers import EulerDiscreteScheduler, StableDiffusionXLPipeline

    extra_kwargs = {}
    if custom_pipeline is not None:
        extra_kwargs["custom_pipeline"] = custom_pipeline
    if variant is not None:
        extra_kwargs["variant"] = variant
    if controlnet is not None:
        from diffusers import ControlNetModel

        controlnet = ControlNetModel.from_pretrained(
            controlnet,
            torch_dtype=torch.float16,
        )
        extra_kwargs["controlnet"] = controlnet
    if os.path.exists(os.path.join(model_name, "calibrate_info.txt")):
        from onediff.quantization import QuantPipeline

        raise TypeError("Quantizatble SDXL-LIGHT is not supported!")
        # pipe = QuantPipeline.from_quantized(
        #     pipeline_cls, model_name, torch_dtype=torch.float16, **extra_kwargs
        # )
    else:
        is_lora_cpkt = "lora" in cpkt_name
        pipe = StableDiffusionXLPipeline.from_pretrained(
            model_name, torch_dtype=torch.float16, **extra_kwargs
        )

        if is_lora_cpkt:
            pipe = StableDiffusionXLPipeline.from_pretrained(
                model_name, torch_dtype=torch.float16, **extra_kwargs
            ).to("cuda")
            if os.path.isfile(os.path.join(repo_name, cpkt_name)):
                pipe.load_lora_weights(os.path.join(repo_name, cpkt_name))
            else:
                pipe.load_lora_weights(hf_hub_download(repo_name, cpkt_name))
            pipe.fuse_lora()
        else:
            from diffusers import UNet2DConditionModel

            unet = UNet2DConditionModel.from_config(model_name, subfolder="unet").to(
                "cuda", torch.float16
            )
            if os.path.isfile(os.path.join(repo_name, cpkt_name)):
                unet.load_state_dict(
                    load_file(os.path.join(repo_name, cpkt_name), device="cuda")
                )
            else:
                from huggingface_hub import hf_hub_download

                unet.load_state_dict(
                    load_file(hf_hub_download(repo_name, cpkt_name), device="cuda")
                )
            pipe = StableDiffusionXLPipeline.from_pretrained(
                model_name, unet=unet, torch_dtype=torch.float16, **extra_kwargs
            ).to("cuda")

    pipe.scheduler = EulerDiscreteScheduler.from_config(
        pipe.scheduler.config, timestep_spacing="trailing"
    )
    pipe.safety_checker = None
    pipe.to(torch.device("cuda"))

    if compile_type == "none":
        pass
    elif compile_type == "oneflow":
        pipe = compile_pipe(pipe)
    elif compile_type in ("compile", "compile-max-autotune"):
        mode = "max-autotune" if compile_type == "compile-max-autotune" else None
        pipe.unet = torch.compile(pipe.unet, mode=mode)
        if hasattr(pipe, "controlnet"):
            pipe.controlnet = torch.compile(pipe.controlnet, mode=mode)
        pipe.vae = torch.compile(pipe.vae, mode=mode)
    else:
        raise ValueError(f"Unknown compiler: {compile_type}")

    return pipe


class IterationProfiler:
    def __init__(self):
        self.begin = None
        self.end = None
        self.num_iterations = 0

    def get_iter_per_sec(self):
        if self.begin is None or self.end is None:
            return None
        self.end.synchronize()
        dur = self.begin.elapsed_time(self.end)
        return self.num_iterations / dur * 1000.0

    def callback_on_step_end(self, pipe, i, t, callback_kwargs={}):
        if self.begin is None:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.begin = event
        else:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.end = event
            self.num_iterations += 1
        return callback_kwargs


def main():
    args = parse_args()

    pipe = load_and_compile_pipe(
        args.model,
        args.repo,
        args.cpkt,
        args.compiler,
        variant=args.variant,
        custom_pipeline=args.custom_pipeline,
        controlnet=args.controlnet,
    )

    height = args.height
    width = args.width
    height = args.height or pipe.unet.config.sample_size * pipe.vae_scale_factor
    width = args.width or pipe.unet.config.sample_size * pipe.vae_scale_factor

    n_steps = int(args.cpkt[len("sdxl_lightning_") : len("sdxl_lightning_") + 1])

    def get_kwarg_inputs():
        kwarg_inputs = dict(
            prompt=args.prompt,
            negative_prompt=args.negative_prompt,
            height=height,
            width=width,
            num_inference_steps=n_steps,
            num_images_per_prompt=args.batch,
            generator=None
            if args.seed is None
            else torch.Generator(device="cuda").manual_seed(args.seed),
            **(
                dict()
                if args.extra_call_kwargs is None
                else json.loads(args.extra_call_kwargs)
            ),
        )
        return kwarg_inputs

    # NOTE: Warm it up.
    # The initial calls will trigger compilation and might be very slow.
    # After that, it should be very fast.
    if args.warmups > 0:
        print("Begin warmup")
        for _ in range(args.warmups):
            pipe(**get_kwarg_inputs())
        print("End warmup")

    # Let"s see it!
    # Note: Progress bar might work incorrectly due to the async nature of CUDA.
    kwarg_inputs = get_kwarg_inputs()
    iter_profiler = None
    if "callback_on_step_end" in inspect.signature(pipe).parameters:
        iter_profiler = IterationProfiler()
        kwarg_inputs["callback_on_step_end"] = iter_profiler.callback_on_step_end
    elif "callback" in inspect.signature(pipe).parameters:
        iter_profiler = IterationProfiler()
        kwarg_inputs["callback"] = iter_profiler.callback_on_step_end
    begin = time.time()
    output_images = pipe(**kwarg_inputs).images
    end = time.time()

    print("=======================================")
    print(f"Inference time: {end - begin:.3f}s")
    iter_per_sec = iter_profiler.get_iter_per_sec()
    if iter_per_sec is not None:
        print(f"Iterations per second: {iter_per_sec:.3f}")
    cuda_mem_after_used = flow._oneflow_internal.GetCUDAMemoryUsed()
    host_mem_after_used = flow._oneflow_internal.GetCPUMemoryUsed()
    print(f"CUDA Mem after: {cuda_mem_after_used / 1024:.3f}GiB")
    print(f"Host Mem after: {host_mem_after_used / 1024:.3f}GiB")
    print("=======================================")

    if args.output_image is not None:
        output_images[0].save(args.output_image)
    else:
        print("Please set `--output-image` to save the output image")


if __name__ == "__main__":
    main()



""" ComfyUI/comfy/ldm/modules/diffusionmodules/util.py """
import oneflow as torch
import oneflow.nn as nn
import oneflow.nn.functional as F
from einops import rearrange


class AlphaBlender(nn.Module):
    strategies = ["learned", "fixed", "learned_with_images"]

    def get_alpha(self, image_only_indicator: torch.Tensor, device) -> torch.Tensor:
        # skip_time_mix = rearrange(repeat(skip_time_mix, 'b -> (b t) () () ()', t=t), '(b t) 1 ... -> b 1 t ...', t=t)
        if self.merge_strategy == "fixed":
            # make shape compatible
            # alpha = repeat(self.mix_factor, '1 -> b () t  () ()', t=t, b=bs)
            alpha = self.mix_factor.to(device)
        elif self.merge_strategy == "learned":
            alpha = torch.sigmoid(self.mix_factor.to(device))
            # make shape compatible
            # alpha = repeat(alpha, '1 -> s () ()', s = t * bs)
        elif self.merge_strategy == "learned_with_images":
            # assert image_only_indicator is not None, "need image_only_indicator ..."
            if image_only_indicator is None:
                # alpha = rearrange(
                #     torch.sigmoid(self.mix_factor.to(device)), "... -> ... 1"
                # )
                # Rewrite for onediff SVD dynamic shape
                alpha = torch.sigmoid(self.mix_factor.to(device)).unsqueeze(-1)
            else:
                # alpha = torch.where(
                #     image_only_indicator.bool(),
                #     torch.ones(1, 1, device=image_only_indicator.device),
                #     rearrange(torch.sigmoid(self.mix_factor.to(image_only_indicator.device)), "... -> ... 1"),
                # )
                # Rewrite for onediff SVD dynamic shape
                alpha = torch.where(
                    image_only_indicator.bool(),
                    torch.ones(1, 1, device=image_only_indicator.device),
                    torch.sigmoid(
                        self.mix_factor.to(image_only_indicator.device)
                    ).unsqueeze(-1),
                )
            # alpha = rearrange(alpha, self.rearrange_pattern)
            # Rewrite for onediff SVD dynamic shape, only VideoResBlock, rearrange_pattern="b t -> b 1 t 1 1",
            if self.rearrange_pattern == "b t -> b 1 t 1 1":
                alpha.unsqueeze(1).unsqueeze(3).unsqueeze(4)
            elif self.rearrange_pattern == "b t -> (b t) 1 1":
                alpha.flatten().unsqueeze(-1).unsqueeze(-1)
            else:
                alpha = rearrange(alpha, self.rearrange_pattern)
            # make shape compatible
            # alpha = repeat(alpha, '1 -> s () ()', s = t * bs)
        else:
            raise NotImplementedError()
        return alpha

    def forward(
        self,
        x_spatial,
        x_temporal,
        image_only_indicator=None,
    ) -> torch.Tensor:
        alpha = self.get_alpha(image_only_indicator, x_spatial.device)
        x = (
            alpha.to(x_spatial.dtype) * x_spatial
            + (1.0 - alpha).to(x_spatial.dtype) * x_temporal
        )
        return x



"""oneflow/framework/args_tree.py
fix: TypeError: can only concatenate str (not "tuple") to str
TODO: fix in oneflow
"""
import oneflow as flow  # usort: skip
from oneflow.framework.args_tree import NamedArg


class PatchNamedArg(NamedArg):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        name = getattr(self, "_name", None)
        setattr(self, "_name", str(name) if name is not None else None)


flow.framework.args_tree.NamedArg = PatchNamedArg



from modules.sd_vae_approx import VAEApprox

from onediff.infer_compiler.backends.oneflow.transform import proxy_class


# Prevent re-importing modules.shared, which incorrectly initializes all its variables.
class VAEApproxOflow(proxy_class(VAEApprox)):
    pass


torch2oflow_class_map = {
    VAEApprox: VAEApproxOflow,
}



import os
import traceback

COMFYUI_ROOT = os.getenv("COMFYUI_ROOT")
from onediff.infer_compiler.backends.oneflow.import_tools import DynamicModuleLoader
from onediff.infer_compiler.backends.oneflow.transform import transform_mgr

from ...sd_hijack_utils import Hijacker

__all__ = ["ipadapter_plus_pt", "ipadapter_plus_of"]

pkg_name = "ComfyUI_IPAdapter_plus"
pkg_root = os.path.join(COMFYUI_ROOT, "custom_nodes", pkg_name)
is_load_ipadapter_plus_pkg = True
try:
    if os.path.exists(pkg_root):
        ipadapter_plus_pt = DynamicModuleLoader.from_path(pkg_root)
        ipadapter_plus_of = transform_mgr.transform_package(pkg_name)
    else:
        is_load_ipadapter_plus_pkg = False
except Exception as e:
    print(traceback.format_exc())
    print(f"Warning: Failed to load {pkg_root} due to {e}")
    is_load_ipadapter_plus_pkg = False
ipadapter_plus_hijacker = Hijacker()



from functools import partial

from ..booster_utils import is_using_nexfort_backend

from ..hijack_ipadapter_plus.set_model_patch_replace import set_model_patch_replace
from ._config import pulid_comfyui, pulid_comfyui_hijacker

# ComfyUI/custom_nodes/PuLID_ComfyUI/pulid.py
set_model_patch_replace_fn = pulid_comfyui.pulid.set_model_patch_replace
pulid_attention = pulid_comfyui.pulid.pulid_attention


set_model_patch_replace_puild = partial(
    set_model_patch_replace, attention_func=pulid_attention
)


def cond_func(org_fn, model, *args, **kwargs):
    return is_using_nexfort_backend(model)


pulid_comfyui_hijacker.register(
    set_model_patch_replace_fn, set_model_patch_replace_puild, cond_func
)



import launch


def install():
    if not launch.is_installed("oneflow"):
        print("oneflow is not installed! Installing...")
        launch.run_pip(
            "install --pre oneflow -f https://oneflow-pro.oss-cn-beijing.aliyuncs.com/branch/community/cu118"
        )
    if not launch.is_installed("onediff"):
        print("onediff is not installed! Installing...")
        launch.run_pip("install git+https://github.com/siliconflow/onediff.git")


install()



from functools import wraps

import onediff_shared
from compile import get_compiled_graph
from compile.utils import is_nexfort_backend, is_oneflow_backend

from .hijack import hijack_controlnet_extension
from .utils import check_if_controlnet_enabled


def onediff_controlnet_decorator(func):
    @wraps(func)
    # TODO: restore hijacked func here
    def wrapper(self, p, *arg, **kwargs):
        try:
            onediff_shared.controlnet_enabled = check_if_controlnet_enabled(p)
            if onediff_shared.controlnet_enabled:
                hijack_controlnet_extension(p)
            return func(self, p, *arg, **kwargs)
        finally:
            if onediff_shared.controlnet_enabled:
                onediff_shared.previous_is_controlnet = True
            else:
                onediff_shared.controlnet_compiled = False
                onediff_shared.previous_is_controlnet = False

    return wrapper


def compile_controlnet_ldm_unet(sd_model, unet_model, *, backend=None, options=None):
    if is_oneflow_backend():
        from compile.oneflow.mock.controlnet import OneFlowOnediffControlNetModel

        from onediff.infer_compiler.backends.oneflow.transform import register

        from .model import OnediffControlNetModel

        register(
            package_names=["scripts.hook"],
            torch2oflow_class_map={
                OnediffControlNetModel: OneFlowOnediffControlNetModel,
            },
        )
    elif is_nexfort_backend():
        # nothing need to do
        pass
    compiled_graph = get_compiled_graph(
        sd_model, unet_model, backend=backend, options=options
    )
    compiled_graph.name += "_controlnet"
    return compiled_graph



import copy
from functools import singledispatchmethod, wraps
from typing import List

import torch.nn as nn
from comfy import model_management
from comfy.model_patcher import ModelPatcher
from comfy.sd import VAE

from .booster_cache import BoosterCacheService
from .booster_interface import BoosterExecutor, BoosterSettings


def auto_cache_model(func):
    @wraps(func)
    def wrapper(self: "BoosterScheduler", model=None, *args, **kwargs):
        if self.settings is None:
            return func(self, model, *args, **kwargs)
        cached_model_key = self.settings.tmp_cache_key
        cached_model = self.cache_service.get_cached_model(cached_model_key, model)
        if cached_model is not None:
            return cached_model
        cached_model = func(self, model, *args, **kwargs)
        self.cache_service.put(cached_model_key, cached_model)
        return cached_model

    return wrapper


class BoosterScheduler:
    def __init__(
        self,
        booster_executors: List[BoosterExecutor],
        *,
        inplace=True,
        settings: BoosterSettings = None,
    ):
        if not isinstance(booster_executors, (list, tuple)):
            booster_executors = [booster_executors]
        self.booster_executors = booster_executors
        self.inplace = inplace
        self.settings = settings
        self.cache_service = BoosterCacheService()

    def is_empty(self) -> bool:
        """
        Checks if the list of boosters is empty.
        """
        return not self.booster_executors

    @auto_cache_model
    def compile(self, model=None, ckpt_name=None, **kwargs):
        if not self.inplace:
            model = self.copy(model)
        for executor in self.booster_executors:
            model = executor.execute(model, ckpt_name=ckpt_name, **kwargs)

        return model

    def __call__(self, model=None, ckpt_name=None, **kwargs):
        return self.compile(model=model, ckpt_name=ckpt_name, **kwargs)

    @singledispatchmethod
    def copy(self, model):
        raise NotImplementedError(f"Copying {type(model)} is not implemented.")

    @copy.register
    def _(self, model: ModelPatcher):
        model.model = model.model.to("cpu")
        new_modelpatcher = model.clone()
        copied_model: nn.Module = copy.deepcopy(model.model)
        new_modelpatcher.model = copied_model.to(model_management.get_torch_device())
        return new_modelpatcher

    @copy.register
    def _(self, model: VAE):
        new_vae = copy.deepcopy(model)
        return new_vae



from typing import Any, Dict, Optional

import torch
from onediff.infer_compiler.backends.oneflow.transform import transform_mgr

transformed_diffusers = transform_mgr.transform_package("diffusers")
ConfigMixin = transformed_diffusers.configuration_utils.ConfigMixin
register_to_config = transformed_diffusers.configuration_utils.register_to_config
ImagePositionalEmbeddings = (
    transformed_diffusers.models.embeddings.ImagePositionalEmbeddings
)
USE_PEFT_BACKEND = transformed_diffusers.utils.USE_PEFT_BACKEND
BaseOutput = transformed_diffusers.utils.BaseOutput
deprecate = transformed_diffusers.utils.deprecate
is_torch_version = transformed_diffusers.utils.is_torch_version
BasicTransformerBlock = transformed_diffusers.models.attention.BasicTransformerBlock
PatchEmbed = transformed_diffusers.models.embeddings.PatchEmbed
PixArtAlphaTextProjection = (
    transformed_diffusers.models.embeddings.PixArtAlphaTextProjection
)
LoRACompatibleConv = transformed_diffusers.models.lora.LoRACompatibleConv
LoRACompatibleLinear = transformed_diffusers.models.lora.LoRACompatibleLinear
ModelMixin = transformed_diffusers.models.modeling_utils.ModelMixin
AdaLayerNormSingle = transformed_diffusers.models.normalization.AdaLayerNormSingle
Transformer2DModelOutput = (
    transformed_diffusers.models.transformers.transformer_2d.Transformer2DModelOutput
)
proxy_Transformer2DModel = (
    transformed_diffusers.models.transformers.transformer_2d.Transformer2DModel
)


class Transformer2DModel(proxy_Transformer2DModel):
    def forward(
        self,
        hidden_states: torch.Tensor,
        encoder_hidden_states: Optional[torch.Tensor] = None,
        timestep: Optional[torch.LongTensor] = None,
        added_cond_kwargs: Dict[str, torch.Tensor] = None,
        class_labels: Optional[torch.LongTensor] = None,
        cross_attention_kwargs: Dict[str, Any] = None,
        attention_mask: Optional[torch.Tensor] = None,
        encoder_attention_mask: Optional[torch.Tensor] = None,
        return_dict: bool = True,
    ):
        # import ipdb; ipdb.set_trace()
        """
        The [`Transformer2DModel`] forward method.

        Args:
            hidden_states (`torch.LongTensor` of shape `(batch size, num latent pixels)` if discrete, `torch.Tensor` of shape `(batch size, channel, height, width)` if continuous):
                Input `hidden_states`.
            encoder_hidden_states ( `torch.Tensor` of shape `(batch size, sequence len, embed dims)`, *optional*):
                Conditional embeddings for cross attention layer. If not given, cross-attention defaults to
                self-attention.
            timestep ( `torch.LongTensor`, *optional*):
                Used to indicate denoising step. Optional timestep to be applied as an embedding in `AdaLayerNorm`.
            class_labels ( `torch.LongTensor` of shape `(batch size, num classes)`, *optional*):
                Used to indicate class labels conditioning. Optional class labels to be applied as an embedding in
                `AdaLayerZeroNorm`.
            cross_attention_kwargs ( `Dict[str, Any]`, *optional*):
                A kwargs dictionary that if specified is passed along to the `AttentionProcessor` as defined under
                `self.processor` in
                [diffusers.models.attention_processor](https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/attention_processor.py).
            attention_mask ( `torch.Tensor`, *optional*):
                An attention mask of shape `(batch, key_tokens)` is applied to `encoder_hidden_states`. If `1` the mask
                is kept, otherwise if `0` it is discarded. Mask will be converted into a bias, which adds large
                negative values to the attention scores corresponding to "discard" tokens.
            encoder_attention_mask ( `torch.Tensor`, *optional*):
                Cross-attention mask applied to `encoder_hidden_states`. Two formats supported:

                    * Mask `(batch, sequence_length)` True = keep, False = discard.
                    * Bias `(batch, 1, sequence_length)` 0 = keep, -10000 = discard.

                If `ndim == 2`: will be interpreted as a mask, then converted into a bias consistent with the format
                above. This bias will be added to the cross-attention scores.
            return_dict (`bool`, *optional*, defaults to `True`):
                Whether or not to return a [`~models.unets.unet_2d_condition.UNet2DConditionOutput`] instead of a plain
                tuple.

        Returns:
            If `return_dict` is True, an [`~models.transformers.transformer_2d.Transformer2DModelOutput`] is returned,
            otherwise a `tuple` where the first element is the sample tensor.
        """
        if cross_attention_kwargs is not None:
            if cross_attention_kwargs.get("scale", None) is not None:
                logger.warning(
                    "Passing `scale` to `cross_attention_kwargs` is deprecated. `scale` will be ignored."
                )
        # ensure attention_mask is a bias, and give it a singleton query_tokens dimension.
        #   we may have done this conversion already, e.g. if we came here via UNet2DConditionModel#forward.
        #   we can tell by counting dims; if ndim == 2: it's a mask rather than a bias.
        # expects mask of shape:
        #   [batch, key_tokens]
        # adds singleton query_tokens dimension:
        #   [batch,                    1, key_tokens]
        # this helps to broadcast it as a bias over attention scores, which will be in one of the following shapes:
        #   [batch,  heads, query_tokens, key_tokens] (e.g. torch sdp attn)
        #   [batch * heads, query_tokens, key_tokens] (e.g. xformers or classic attn)
        if attention_mask is not None and attention_mask.ndim == 2:
            # assume that mask is expressed as:
            #   (1 = keep,      0 = discard)
            # convert mask into a bias that can be added to attention scores:
            #       (keep = +0,     discard = -10000.0)
            attention_mask = (1 - attention_mask.to(hidden_states.dtype)) * -10000.0
            attention_mask = attention_mask.unsqueeze(1)

        # convert encoder_attention_mask to a bias the same way we do for attention_mask
        if encoder_attention_mask is not None and encoder_attention_mask.ndim == 2:
            encoder_attention_mask = (
                1 - encoder_attention_mask.to(hidden_states.dtype)
            ) * -10000.0
            encoder_attention_mask = encoder_attention_mask.unsqueeze(1)

        hidden_states_in = hidden_states

        # 1. Input
        if self.is_input_continuous:
            batch_size, _, height, width = hidden_states.shape
            residual = hidden_states
            hidden_states, inner_dim = self._operate_on_continuous_inputs(hidden_states)
        elif self.is_input_vectorized:
            hidden_states = self.latent_image_embedding(hidden_states)
        elif self.is_input_patches:
            height, width = (
                hidden_states.shape[-2] // self.patch_size,
                hidden_states.shape[-1] // self.patch_size,
            )
            (
                hidden_states,
                encoder_hidden_states,
                timestep,
                embedded_timestep,
            ) = self._operate_on_patched_inputs(
                hidden_states, encoder_hidden_states, timestep, added_cond_kwargs
            )

        # 2. Blocks
        for block in self.transformer_blocks:
            if self.training and self.gradient_checkpointing:

                def create_custom_forward(module, return_dict=None):
                    def custom_forward(*inputs):
                        if return_dict is not None:
                            return module(*inputs, return_dict=return_dict)
                        else:
                            return module(*inputs)

                    return custom_forward

                ckpt_kwargs: Dict[str, Any] = (
                    {"use_reentrant": False} if is_torch_version(">=", "1.11.0") else {}
                )
                hidden_states = torch.utils.checkpoint.checkpoint(
                    create_custom_forward(block),
                    hidden_states,
                    attention_mask,
                    encoder_hidden_states,
                    encoder_attention_mask,
                    timestep,
                    cross_attention_kwargs,
                    class_labels,
                    **ckpt_kwargs,
                )
            else:
                hidden_states = block(
                    hidden_states,
                    attention_mask=attention_mask,
                    encoder_hidden_states=encoder_hidden_states,
                    encoder_attention_mask=encoder_attention_mask,
                    timestep=timestep,
                    cross_attention_kwargs=cross_attention_kwargs,
                    class_labels=class_labels,
                )

        # 3. Output
        if self.is_input_continuous:
            if not self.use_linear_projection:
                hidden_states = (
                    hidden_states.permute(0, 2, 1)
                    .reshape_as(hidden_states_in)
                    .contiguous()
                )
                # hidden_states = (
                #     hidden_states.reshape(batch_size, height, width, inner_dim).permute(0, 3, 1, 2).contiguous()
                # )
                hidden_states = self.proj_out(hidden_states)
            else:
                hidden_states = self.proj_out(hidden_states)
                hidden_states = (
                    hidden_states.permute(0, 2, 1)
                    .reshape_as(hidden_states_in)
                    .contiguous()
                )
                # hidden_states = (
                #     hidden_states.reshape(batch_size, height, width, inner_dim).permute(0, 3, 1, 2).contiguous()
                # )

            output = hidden_states + residual
            # output = self._get_output_for_continuous_inputs(
            #     hidden_states=hidden_states,
            #     residual=residual,
            #     batch_size=batch_size,
            #     height=height,
            #     width=width,
            #     inner_dim=inner_dim,
            # )
        elif self.is_input_vectorized:
            output = self._get_output_for_vectorized_inputs(hidden_states)
        elif self.is_input_patches:
            output = self._get_output_for_patched_inputs(
                hidden_states=hidden_states,
                timestep=timestep,
                class_labels=class_labels,
                embedded_timestep=embedded_timestep,
                height=height,
                width=width,
            )

        if not return_dict:
            return (output,)

        return Transformer2DModelOutput(sample=output)

    def _get_output_for_continuous_inputs(
        self, hidden_states, residual, batch_size, height, width, inner_dim
    ):
        # # hidden_states = hidden_states.reshape(batch, height, width, inner_dim).permute(0, 3, 1, 2).contiguous()
        # hidden_states = (
        #     hidden_states.permute(0, 2, 1)
        #     .reshape_as(hidden_states_in)
        #     .contiguous()
        # )
        # hidden_states = self.proj_out(hidden_states)
        # import ipdb; ipdb.set_trace()
        return
        if not self.use_linear_projection:
            hidden_states = (
                hidden_states.reshape(batch_size, height, width, inner_dim)
                .permute(0, 3, 1, 2)
                .contiguous()
            )
            hidden_states = self.proj_out(hidden_states)
        else:
            hidden_states = self.proj_out(hidden_states)
            hidden_states = (
                hidden_states.reshape(batch_size, height, width, inner_dim)
                .permute(0, 3, 1, 2)
                .contiguous()
            )

        output = hidden_states + residual
        return output

    def _get_output_for_vectorized_inputs(self, hidden_states):
        raise
        hidden_states = self.norm_out(hidden_states)
        logits = self.out(hidden_states)
        # (batch, self.num_vector_embeds - 1, self.num_latent_pixels)
        logits = logits.permute(0, 2, 1)
        # log(p(x_0))
        output = F.log_softmax(logits.double(), dim=1).float()
        return output

    def _get_output_for_patched_inputs(
        self,
        hidden_states,
        timestep,
        class_labels,
        embedded_timestep,
        height=None,
        width=None,
    ):
        raise
        # import ipdb; ipdb.set_trace()
        if self.config.norm_type != "ada_norm_single":
            conditioning = self.transformer_blocks[0].norm1.emb(
                timestep, class_labels, hidden_dtype=hidden_states.dtype
            )
            shift, scale = self.proj_out_1(F.silu(conditioning)).chunk(2, dim=1)
            hidden_states = (
                self.norm_out(hidden_states) * (1 + scale[:, None]) + shift[:, None]
            )
            hidden_states = self.proj_out_2(hidden_states)
        elif self.config.norm_type == "ada_norm_single":
            shift, scale = (
                self.scale_shift_table[None] + embedded_timestep[:, None]
            ).chunk(2, dim=1)
            hidden_states = self.norm_out(hidden_states)
            # Modulation
            hidden_states = hidden_states * (1 + scale) + shift
            hidden_states = self.proj_out(hidden_states)
            hidden_states = hidden_states.squeeze(1)

        # unpatchify
        if self.adaln_single is None:
            height = width = int(hidden_states.shape[1] ** 0.5)
        hidden_states = hidden_states.reshape(
            shape=(
                -1,
                height,
                width,
                self.patch_size,
                self.patch_size,
                self.out_channels,
            )
        )
        hidden_states = torch.einsum("nhwpqc->nchpwq", hidden_states)
        output = hidden_states.reshape(
            shape=(
                -1,
                self.out_channels,
                height * self.patch_size,
                width * self.patch_size,
            )
        )
        return output

    def _operate_on_continuous_inputs(self, hidden_states):
        batch, _, height, width = hidden_states.shape
        hidden_states = self.norm(hidden_states)

        if not self.use_linear_projection:
            hidden_states = self.proj_in(hidden_states)
            inner_dim = hidden_states.shape[1]
            # hidden_states = hidden_states.permute(0, 2, 3, 1).reshape(batch, height * width, inner_dim)
            hidden_states = hidden_states.permute(0, 2, 3, 1).flatten(1, 2)
        else:
            inner_dim = hidden_states.shape[1]
            # hidden_states = hidden_states.permute(0, 2, 3, 1).reshape(batch, height * width, inner_dim)
            hidden_states = hidden_states.permute(0, 2, 3, 1).flatten(1, 2)
            hidden_states = self.proj_in(hidden_states)

        return hidden_states, inner_dim



import json
import threading
import time
import urllib.parse
import urllib.request
import uuid
from contextlib import contextmanager

import websocket  # NOTE: websocket-client (https://github.com/websocket-client/websocket-client)

from .log_utils import logger

__all__ = ["comfy_client_context", "ComfyGraph", "ComfyClient"]


class ComfyGraph:
    def __init__(
        self,
        graph: dict,
        sampler_nodes: list[str],
    ):
        self.graph = graph
        self.sampler_nodes = sampler_nodes

    def set_prompt(self, prompt, negative_prompt=None):
        # Sets the prompt for the sampler nodes (eg. base and refiner)
        for node in self.sampler_nodes:
            prompt_node = self.graph[node]["inputs"]["positive"][0]
            self.graph[prompt_node]["inputs"]["text"] = prompt
            if negative_prompt:
                negative_prompt_node = self.graph[node]["inputs"]["negative"][0]
                self.graph[negative_prompt_node]["inputs"]["text"] = negative_prompt

    def set_sampler_name(
        self,
        sampler_name: str,
    ):
        # sets the sampler name for the sampler nodes (eg. base and refiner)
        for node in self.sampler_nodes:
            self.graph[node]["inputs"]["sampler_name"] = sampler_name

    def set_scheduler(self, scheduler: str):
        # sets the sampler name for the sampler nodes (eg. base and refiner)
        for node in self.sampler_nodes:
            self.graph[node]["inputs"]["scheduler"] = scheduler

    def set_filename_prefix(self, prefix: str):
        # sets the filename prefix for the save nodes
        for node in self.graph:
            if self.graph[node]["class_type"] == "SaveImage":
                self.graph[node]["inputs"]["filename_prefix"] = prefix

    def set_image_size(self, height: int, width: int, batch_size: int = 1):
        for node in self.sampler_nodes:
            size_node = self.graph[node]["inputs"]["latent_image"][0]
            self.graph[size_node]["inputs"]["height"] = height
            self.graph[size_node]["inputs"]["width"] = width
            self.graph[size_node]["inputs"]["batch_size"] = batch_size


@contextmanager
def comfy_client_context(
    listen="127.0.0.1",
    port=30000,
    client_id=str(uuid.uuid4()),
    n_tries=5,
    *args,
    **kwargs,
):
    client = ComfyClient()
    try:
        client.connect(
            listen=listen,
            port=port,
            client_id=client_id,
            n_tries=n_tries,
            *args,
            **kwargs,
        )
        yield client
    except Exception as e:
        logger.error(f"Failed to connect to ComfyClient: {e}")
        raise
    finally:
        logger.info("Closing connection to ComfyClient")
        client.close()


class ComfyClient:
    def connect(
        self, listen="127.0.0.1", port=30000, client_id=str(uuid.uuid4()), n_tries=5
    ):
        for i in range(n_tries):
            time.sleep(4)
            self.client_id = client_id
            self.server_address = f"{listen}:{port}"
            try:
                ws = websocket.WebSocket()
                ws.connect(f"ws://{self.server_address}/ws?clientId={self.client_id}")
                self.ws = ws
            except ConnectionRefusedError as e:
                print(e)
                print(f"({i+1}/{n_tries}) Retrying...")
            else:
                logger.info(f"Connected to server: {self.server_address}")
                threading.Thread(
                    target=self.fetch_system_stats_periodically, daemon=True
                ).start()
                break
        if not self.ws:
            raise RuntimeError(f"Could not connect to server: {self.server_address}")

    def fetch_system_stats_periodically(self):
        max_vram_used_gb = 0
        last_print_time = time.time()
        BYTES_TO_GB = 1024 * 1024 * 1024
        poll_interval = 0.5  # sec
        while True:
            try:
                stats = self.get_system_stats()
                for device in stats["devices"]:
                    vram_total = device["vram_total"]
                    vram_free = device["vram_free"]
                    vram_used = vram_total - vram_free

                    vram_used_gb = vram_used / BYTES_TO_GB

                    if vram_used_gb > max_vram_used_gb:
                        max_vram_used_gb = vram_used_gb

                current_time = time.time()
                if current_time - last_print_time >= 2:
                    logger.info(
                        f"Current VRAM used: {vram_used_gb:.2f} GB\tMaximum VRAM used: {max_vram_used_gb:.2f} GB"
                    )
                    last_print_time = current_time

                time.sleep(poll_interval)
            except Exception as e:
                print(f"{e=}")
                break

    def close(self):
        self.ws.close()

    def queue_prompt(self, prompt):
        p = {"prompt": prompt, "client_id": self.client_id}
        data = json.dumps(p).encode("utf-8")
        req = urllib.request.Request(f"http://{self.server_address}/prompt", data=data)
        return json.loads(urllib.request.urlopen(req).read())

    def get_image(self, filename, subfolder, folder_type):
        data = {"filename": filename, "subfolder": subfolder, "type": folder_type}
        url_values = urllib.parse.urlencode(data)
        with urllib.request.urlopen(
            f"http://{self.server_address}/view?{url_values}"
        ) as response:
            return response.read()

    def get_history(self, prompt_id):
        with urllib.request.urlopen(
            f"http://{self.server_address}/history/{prompt_id}"
        ) as response:
            return json.loads(response.read())

    def get_system_stats(self):
        with urllib.request.urlopen(
            f"http://{self.server_address}/system_stats"
        ) as response:
            return json.loads(response.read())

    def get_images(self, graph, save=True):
        prompt = graph.graph if isinstance(graph, ComfyGraph) else graph

        if not save:
            prompt_str = json.dumps(prompt)
            prompt_str = prompt_str.replace("SaveImage", "PreviewImage")
            prompt = json.loads(prompt_str)

        prompt_id = self.queue_prompt(prompt)["prompt_id"]
        output_images = {}
        while True:
            out = self.ws.recv()
            if isinstance(out, str):
                message = json.loads(out)
                if message["type"] == "executing":
                    data = message["data"]
                    if data["node"] is None and data["prompt_id"] == prompt_id:
                        break
            else:
                continue

        history = self.get_history(prompt_id)[prompt_id]
        for o in history["outputs"]:
            for node_id in history["outputs"]:
                node_output = history["outputs"][node_id]
                if "images" in node_output:
                    images_output = []
                    for image in node_output["images"]:
                        image_data = self.get_image(
                            image["filename"], image["subfolder"], image["type"]
                        )
                        images_output.append(image_data)
                    output_images[node_id] = images_output

        return output_images



import torch
from comfy import model_management
from comfy.model_base import SVD_img2vid
from onediff.infer_compiler import oneflow_compile
from register_comfy import DeepCacheUNet, FastDeepCacheUNet

from .booster_utils import set_environment_for_svd_img2vid

from .model_patcher import OneFlowDeepCacheSpeedUpModelPatcher


def deep_cache_speedup(
    model,
    use_graph,
    cache_interval,
    cache_layer_id,
    cache_block_id,
    start_step,
    end_step,
    *,
    gen_compile_options=None,
    use_oneflow_deepcache_speedup_modelpatcher=True,
):

    offload_device = model_management.unet_offload_device()
    if use_oneflow_deepcache_speedup_modelpatcher:
        model_patcher = OneFlowDeepCacheSpeedUpModelPatcher(
            model.model,
            load_device=model_management.get_torch_device(),
            offload_device=offload_device,
            cache_layer_id=cache_layer_id,
            cache_block_id=cache_block_id,
            use_graph=use_graph,
            gen_compile_options=gen_compile_options,
        )
    else:
        model_patcher = model
        model_patcher.deep_cache_unet = DeepCacheUNet(
            model_patcher.model.diffusion_model, cache_layer_id, cache_block_id
        )
        model_patcher.fast_deep_cache_unet = FastDeepCacheUNet(
            model_patcher.model.diffusion_model, cache_layer_id, cache_block_id
        )
        model_patcher.deep_cache_unet = oneflow_compile(model_patcher.deep_cache_unet)
        model_patcher.fast_deep_cache_unet = oneflow_compile(
            model_patcher.fast_deep_cache_unet
        )

    current_t = -1
    current_step = -1
    cache_h = None

    _first_run = True

    def apply_model(model_function, kwargs):
        if isinstance(model_patcher.model, SVD_img2vid):
            set_environment_for_svd_img2vid(model_patcher)
        nonlocal current_t, current_step, cache_h, _first_run

        if _first_run:
            if hasattr(model_patcher.deep_cache_unet, "quantize"):
                model_patcher.deep_cache_unet.quantize()

            if hasattr(model_patcher.fast_deep_cache_unet, "quantize"):
                model_patcher.fast_deep_cache_unet.quantize()
            _first_run = False

        xa = kwargs["input"]
        t = kwargs["timestep"]
        c_concat = kwargs["c"].get("c_concat", None)
        c_crossattn = kwargs["c"].get("c_crossattn", None)
        y = kwargs["c"].get("y", None)
        control = kwargs["c"].get("control", None)
        transformer_options = kwargs["c"].get("transformer_options", None)

        # https://github.com/comfyanonymous/ComfyUI/blob/629e4c552cc30a75d2756cbff8095640af3af163/comfy/model_base.py#L51-L69
        sigma = t
        xc = model_patcher.model.model_sampling.calculate_input(sigma, xa)
        if c_concat is not None:
            xc = torch.cat([xc] + [c_concat], dim=1)

        context = c_crossattn
        dtype = model_patcher.model.get_dtype()
        xc = xc.to(dtype)
        t = model_patcher.model.model_sampling.timestep(t).float()
        context = context.to(dtype)
        extra_conds = {}
        for o in kwargs:
            extra = kwargs[o]
            if hasattr(extra, "to"):
                extra = extra.to(dtype)
            extra_conds[o] = extra

        x = xc
        timesteps = t
        y = None if y is None else y.to(dtype)
        transformer_options["original_shape"] = list(x.shape)
        transformer_options["current_index"] = 0
        transformer_patches = transformer_options.get("patches", {})
        """
            Apply the model to an input batch.
            :param x: an [N x C x ...] Tensor of inputs.
            :param timesteps: a 1-D batch of timesteps.
            :param context: conditioning plugged in via crossattn
            :param y: an [N] Tensor of labels, if class-conditional.
            :return: an [N x C x ...] Tensor of outputs.
            """

        # reference https://gist.github.com/laksjdjf/435c512bc19636e9c9af4ee7bea9eb86
        if t[0].item() > current_t:
            current_step = -1

        current_t = t[0].item()
        apply = 1000 - end_step <= current_t <= 1000 - start_step  # t is 999->0

        if apply:
            current_step += 1
        else:
            current_step = -1
        current_t = t[0].item()

        is_slow_step = current_step % cache_interval == 0 and apply

        model_output = None
        if is_slow_step:
            cache_h = None
            model_output, cache_h = model_patcher.deep_cache_unet(
                x,
                timesteps,
                context,
                y,
                control,
                transformer_options,
                **extra_conds,
            )
        else:
            model_output, cache_h = model_patcher.fast_deep_cache_unet(
                x,
                cache_h,
                timesteps,
                context,
                y,
                control,
                transformer_options,
                **extra_conds,
            )

        return model_patcher.model.model_sampling.calculate_denoised(
            sigma, model_output, xa
        )

    model_patcher.set_model_unet_function_wrapper(apply_model)
    return (model_patcher,)
